from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from .AbstractGLPlotItem import AbstractGLPlotItem
    from ..widgets.GL3DViewWidget import GL3DViewWidget
    from .GL3DGraphicsItemMixin import GL3DGraphicsItemMixin


import weakref
from PySide6 import QtCore, QtGui, QtWidgets
from pyqtgraph.Qt import QT_LIB
from OpenGL import GL, GLU
from pyqtgraph import Point, Vector, opengl
import numpy as np

from .AbstractGLPlotItem import AbstractGLPlotItem
from .GL3DGraphicsItemMixin import GL3DGraphicsItemMixin
from ..meta import QABCMeta
from abc import abstractmethod
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem

import math
import logging
import warnings
from .GL3DItemGroup import GL3DItemGroup
from dataclasses import dataclass, field
import warnings
from pprint import pprint
import pyqtgraph as pg

class WeakList(object):

    def __init__(self):
        self._items = []

    def append(self, obj):
        #Add backwards to iterate backwards (to make iterating more efficient on removal).
        self._items.insert(0, weakref.ref(obj))

    def __iter__(self):
        i = len(self._items)-1
        while i >= 0:
            ref = self._items[i]
            d = ref()
            if d is None:
                del self._items[i]
            else:
                yield d
            i -= 1



@dataclass(slots=True)
class _MouseMetrics:
    initPos: np.ndarray|None = field(default=None) 
    pos: np.ndarray|None = field(default=None) 
    lastPos: np.ndarray|None = field(default=None) 
        
    def setInit(self, *args):
        self.pos = np.array(args) if isinstance(*args, (list, tuple)) else args
        self.pos = self.pos[0] if isinstance(self.pos, tuple) and isinstance(self.pos[0], np.ndarray) else self.pos
        self.initPos = self.pos.copy()

    def update(self, args):
        self.lastPos = self.pos.copy()
        #self.pos[:] = args
        self.pos = args
        
    def reset(self):
        self.pos=None
        self.initPos=None
        self.lastPos=None
        
    def validPos(self):
        return not self.pos is None
    
    def delta(self):
        return self.pos - self.lastPos

    def deltaSincePress(self):
        return self.pos - self.initPos


@dataclass(slots=True)
class MouseState:
    intersectedPoint: _MouseMetrics = field(init=False, default=None)
    window: _MouseMetrics = field(init=False, default=None)
    world: _MouseMetrics = field(init=False, default=None)
    
    _interactionKeys: Dict[str, QtCore.Qt.Key] = field(init=False, default=dict)
    _interactionState: Dict[str, bool] = field(init=False, default=dict)
    _btn_modes: Dict[str, bool] = field(init=False, default=dict)
    _opts: Dict[str, bool] = field(init=False, default=dict)

    def __post_init__(self):
        self.intersectedPoint = _MouseMetrics()
        self.window = _MouseMetrics()
        self.world = _MouseMetrics()
        self._interactionKeys = {"axisInteracting" : QtCore.Qt.KeyboardModifier.ControlModifier,
                                 "constantAxisZoom" : (QtCore.Qt.KeyboardModifier.ControlModifier | QtCore.Qt.KeyboardModifier.ShiftModifier),
                                 "pan" : QtCore.Qt.MouseButton.LeftButton,
                                 "axisScale" : QtCore.Qt.MouseButton.RightButton,
                                 
                                 }
        self._interactionState = {"axisInteracting" : False,
                                  "constantAxisZoom" : False,
                                  "intersectedAxis" : None,
                                  "mouse" : False,
                                  "keys" : False,
                                  }
        
        self._opts = {"interactingModifier" : QtCore.Qt.KeyboardModifier.ControlModifier,
                      "constantAxisKey" : QtCore.Qt.KeyboardModifier.ShiftModifier,
                      "interactingMouseBtns" : (QtCore.Qt.MouseButton.LeftButton | QtCore.Qt.MouseButton.RightButton),
                      "axisPanBtn" :  QtCore.Qt.MouseButton.LeftButton,
                      "axisScaleBtn" : QtCore.Qt.MouseButton.RightButton
                      }
        
        self._btn_modes = {"axisPanBtn" :  QtCore.Qt.MouseButton.LeftButton,
                           "axisScaleBtn" : QtCore.Qt.MouseButton.RightButton
                           } 

    def getModType(self, btn):
        if self.mouseInteracting():
            if self.keyInteracting():
                if self._btn_modes["axisPanBtn"] == btn:
                    return "pan"
                elif self._btn_modes["axisScaleBtn"] == btn:
                    return "scale"
    
    def KeyToggleEvent(self, ev: QtGui.QKeyEvent):
        if ev.modifiers() & self._opts["interactingModifier"]:
            if not self.keyInteracting():
                self._setKeyInteracting(True)
            if ev.modifiers() & self._opts["constantAxisKey"]:
                if not self._interactionState["constantAxisZoom"]:
                    self._interactionState["constantAxisZoom"]=True 
        else:
            self._setKeyInteracting(False)

    def userInteracting(self):
        return self.keyInteracting()
    
    def keyInteracting(self):
        return self._interactionState["keys"]
    
    def _setKeyInteracting(self, flag):
        self._interactionState["keys"]=flag
        if not flag:
            self._interactionKeys["constantAxisZoom"]=False
            
    def constantAxis(self):
        return self._interactionState["constantAxisZoom"]
    
    def _setMouseInteracting(self, flag):
        self._interactionState["mouse"]=flag
        
                    
    def mouseToggleEvent(self, ev):
        if ev.buttons() & self._opts["interactingMouseBtns"]:
            if not self._interactionState["mouse"]:
                self._setMouseInteracting(True)
        else:
            self.window.reset()
            self.world.reset()
            self.intersectedPoint.reset()
            self._interactionState["intersectedAxis"] = None
            
    def windowPos(self):
        return self.window.pos()    

    def mouseInteracting(self):
        return self._interactionState["mouse"]
     
    def mouseMove(self, x, y):
        self.window.update(x, y)
    
    def setWorldPoint(self, point):
        self.world.setInit(point)
        
    def updateWorldPoint(self, point):
        self.world.setInit(point)
        self.world.update(point)
        
    def setIntersectPoint(self, ax, point):
        self.intersectedPoint.setInit(point)
        self._interactionState["intersectedAxis"]=ax
        
    def updateIntersectPoint(self, point):
        self.intersectedPoint.update(point)
    
    def setNoInteractionState(self):
        self._interactionState["axisInteracting"] = False
        self._interactionState["constantAxisZoom"] = False
        self._interactionState["intersectedAxis"] = None

    def validWindowPos(self):
        return self.window.validPos()

    def validWorldPos(self):
        return self.world.validPos()
    
    def validIntersectedPoint(self):
        return self.intersectedPoint.validPos()



            
class GL3DViewBox(GL3DGraphicsItemMixin, GLGraphicsItem):

    sigYRangeChanged = QtCore.Signal(object, object)
    sigXRangeChanged = QtCore.Signal(object, object)
    sigZRangeChanged = QtCore.Signal(object, object)
    sigRangeChanged = QtCore.Signal(object, object, object)
    sigRangeChangedManually = QtCore.Signal(object)
    
    sigStateChanged = QtCore.Signal(object)
    sigTransformChanged = QtCore.Signal(object)
    sigResized = QtCore.Signal(object)
    sigViewChange = QtCore.Signal(object)

    
    XAxis = 0
    YAxis = 1
    ZAxis = 2
    XYZAxes = 3

    
    def __init__(self,
                 parentItem=None,
                 worldRange=np.array([[0., 1.]]*3).tolist(),
                 enableMouse=True,
                 defaultPadding=np.array([0.0]*3),
                 **kwargs
                 ):
        super().__init__(parentItem=parentItem)
        self.scalebycounter=0
        self.setFlag(self.GLGraphicsItemFlag.ItemHasNoContents)
        self._updatingRange=False
        self._matrixNeedsUpdate=True
        self._autoRangeNeedsUpdate=True
        
        self._mouseZoomFactor=0.999

        self._view_changed_timer=QtCore.QTimer()
        self._update_axis_ms=100
        self._update_interval = 50 
        self._last_update_time = 0
        self._view_changed_timer.setSingleShot(True)
        self._view_changed_timer.timeout.connect(self._ProcessViewDrag)
        self.mouse_state = MouseState()
        
        self.mouse_info={"mouseAlongAxisPos" : None,
                         "prevmouseAlongAxisPos" : None,
                         "intersectedAxis" : None
                         }
        self._cartesian_view_coords = np.array([0., 0., 0.])
        self._camera_position = np.array([0., 0., 0.])

        self._name=None
        self._viewRange:List[List[float]]=None
        self._itemBoundsCache = weakref.WeakKeyDictionary()
        
        self.state={"autoRange" : [True]*3,
                    "autoVisibleOnly" : [False]*3,
                    "aspectLocked" : False,
                    "defaultPadding" : defaultPadding,
                    "mouseEnabled": [enableMouse, enableMouse, enableMouse],
                    "targetRange" : [[0,1], [0,1], [0, 1]],
                    "viewRange" : [[0,1], [0,1], [0, 1]],
                    "linkedViews" : [None, None, None],
                    "explicitViewRange" : False,
                    "wheelScaleFactor": -1.0 / 14.0,
                    'limits': { 
                                'xLimits': [-1E307, +1E307],   # Maximum and minimum visible X values
                                'yLimits': [-1E307, +1E307],   # Maximum and minimum visible Y values
                                'zLimits' : [-1E307, +1E307],
                                'xRange': [None, None],   # Maximum and minimum X range
                                'yRange': [None, None],   # Maximum and minimum Y range
                                'zRange': [None, None], 
                                },
                    
                    }
        
        
        self.setWorldRange(worldRange)
        
        self.state.update({"worldSize" : [abs(r[0] - r[1]) for r in self.state["worldRange"]],
                           "vertexCoordinates" : self.get_cube_face(self.state["worldRange"]),
                           })
        
        
        self.addedItems: List[GL3DGraphicsItemMixin]=[]
        
        
        
        self._range_arr = {"worldRange" : np.array(self.state["worldRange"], dtype=float),
                           "viewRange" : np.array(self.state["viewRange"], dtype=float),
                           }
        
        
        self.childGroup = GL3DItemGroup(parentItem=self)
        self.childGroup.itemsChangedListeners.append(self)
    
    def implements(self, interface):
        return interface == 'ViewBox'

    def addItem(self, item, ignoreBounds=False):
        item.setParentItem(self.childGroup)
        if not item in self.addedItems and not ignoreBounds:
            self.addedItems.append(item)          

        self.queueUpdateAutoRange()
        
    def removeItem(self, item):
        item.setParentItem(None)
        if item in self.addedItems:
            self.addedItems.remove(item)
            
        self.queueUpdateAutoRange()

        
    
    def setWorldRange(self, worldRange: List[List[float]]):
        for wr in worldRange:
            if wr[0] > wr[1]:
                raise ValueError(f"worldRange[jdx][0] < worldRange[jdx][1] for each axis")
        
        try:
            shape = np.shape(worldRange)
            if shape != (3, 2):
                raise ValueError(F"worldRange must have shape == (3,2)")
        except:            
            raise ValueError(F"worldRange must have shape == (3,2)")
        self.state["worldRange"] = worldRange
        
    def viewRange(self) -> List[List[np.ndarray]]:
        return [x[:] for x in self.state["viewRange"]]
    
    def viewRangeArr(self) -> List[List[np.ndarray]]:
        self._range_arr["viewRange"][:] = self.viewRange()
        return self._range_arr["viewRange"]

    def worldRange(self) -> List[List[np.ndarray]]:
        return [x[:] for x in self.state["worldRange"]]
    
    def worldRangeArr(self) -> List[List[np.ndarray]]:
        self._range_arr["worldRange"][0, :] = self.state["worldRange"][0][:]
        self._range_arr["worldRange"][1, :] = self.state["worldRange"][1][:]
        self._range_arr["worldRange"][2, :] = self.state["worldRange"][2][:]
        return self._range_arr["worldRange"]

    def targetRange(self) -> List[List[np.ndarray]]:
        return [x[:] for x in self.state["targetRange"]]
    
    def _internal_update(self):
        return QtWidgets.QGraphicsItem.update()

    def innerSceneItem(self):
        return self.childGroup

    def queueUpdateAutoRange(self):
        self._autoRangeNeedsUpdate = True
        self.update()

    def _effectiveLimits(self):
        return (self.state['limits']['xLimits'], self.state['limits']['yLimits'], self.state['limits']['zLimits'])
    
    def updateViewRange(self, forceX=False, forceY=False, forceZ=False):
        viewRange = [self.state['targetRange'][0][:], self.state['targetRange'][1][:], self.state['targetRange'][2][:]]
        changed = [False]*3
        
        aspect = self.state['aspectLocked']  # size ratio / view ratio
        limits = self._effectiveLimits()

        minRng = [self.state['limits']['xRange'][0], self.state['limits']['yRange'][0], self.state['limits']['zRange'][0]]
        maxRng = [self.state['limits']['xRange'][1], self.state['limits']['yRange'][1], self.state['limits']['zRange'][1]]

        for ax in [0, 1, 2]:
            if limits[ax][0] is None and limits[ax][1] is None and minRng[ax] is None and maxRng[ax] is None:
                continue
            # max range cannot be larger than bounds, if they are given
            if limits[ax][0] is not None and limits[ax][1] is not None:
                if maxRng[ax] is not None:
                    maxRng[ax] = min(maxRng[ax], limits[ax][1] - limits[ax][0])
                else:
                    maxRng[ax] = limits[ax][1] - limits[ax][0]

        for ax in [0, 1, 2]:
            range = viewRange[ax][1] - viewRange[ax][0]
            if minRng[ax] is not None and minRng[ax] > range:
                viewRange[ax][1] = viewRange[ax][0] + minRng[ax]
                self.state["targetRange"][ax] = viewRange[ax]
            if maxRng[ax] is not None and maxRng[ax] < range:
                viewRange[ax][1] = viewRange[ax][0] + maxRng[ax]
                self.state["targetRange"][ax] = viewRange[ax]
            if limits[ax][0] is not None and viewRange[ax][0] < limits[ax][0]:
                delta = limits[ax][0] - viewRange[ax][0]
                viewRange[ax][0] += delta
                viewRange[ax][1] += delta
                self.state["targetRange"][ax] = viewRange[ax]
            if limits[ax][1] is not None and viewRange[ax][1] > limits[ax][1]:
                delta = viewRange[ax][1] - limits[ax][1]
                viewRange[ax][0] -= delta
                viewRange[ax][1] -= delta
                self.state["targetRange"][ax] = viewRange[ax]
 
        thresholds = [(viewRange[axis][1] - viewRange[axis][0]) * 1.0e-9 for axis in (0,1,2)]
        changed = []
        for ax in [0,1,2]:
            con_1 = abs(viewRange[ax][0] - self.state["viewRange"][ax][0]) > thresholds[ax]
            con_2 = abs(viewRange[ax][1] - self.state["viewRange"][ax][1]) > thresholds[ax]
            changed.append(any((con_1, con_2)))

        self.state["viewRange"] = viewRange
        if any(changed):
            self._matrixNeedsUpdate = True
            self.update()

            # Inform linked views that the range has changed
            for ax in [0, 1, 2]:
                if not changed[ax]:
                    continue
                link = self.linkedView(ax)
                if link is not None:
                    link.linkedViewChanged(self, ax)

            if changed[0]:
                self.sigXRangeChanged.emit(self, tuple(self.state["viewRange"][0]))
            if changed[1]:
                self.sigYRangeChanged.emit(self, tuple(self.state["viewRange"][1]))
            if changed[2]:
                self.sigZRangeChanged.emit(self, tuple(self.state["viewRange"][2]))
            
            self.sigRangeChanged.emit(self, self.state["viewRange"], changed)
            
    def linkedView(self, ax):
        ## Return the linked view for axis *ax*.
        ## this method _always_ returns either a ViewBox or None.
        v = self.state['linkedViews'][ax]
        if v is None or isinstance(v, str):
            return None
        else:
            return v()  ## dereference weakref pointer. If the reference is dead, this returns None

    
    def itemBoundsChanged(self, item):
        self._itemBoundsCache.pop(item, None)
        if any((self.state['autoRange'][0],
                self.state['autoRange'][1],
                self.state['autoRange'][2]
                )):
            self.queueUpdateAutoRange()
            
    def itemsChanged(self):
        self.queueUpdateAutoRange()
    
    def itemChange(self, change, value):
        
        ret = super().itemChange(change, value)

        if change == self.GLGraphicsItemChange.ItemParentWidgetChange:
            parent_widget = self.parentWidget()
            if not parent_widget is None and hasattr(parent_widget, 'sigPrepareForPaint'):
                parent_widget.sigPrepareForPaint.disconnect(self.prepareForPaint)
        elif change == self.GLGraphicsItemChange.ItemParentWidgetHasChanged:
            
            parent_widget = self.parentWidget()
            if not parent_widget is None and hasattr(parent_widget, 'sigPrepareForPaint'):
                parent_widget.sigPrepareForPaint.connect(self.prepareForPaint)
            parent_widget.sigPrepareForPaint.emit()
        return ret
                        
    def _resetTarget(self):
        self.state['targetRange'] = [self.state["viewRange"][0][:], self.state["viewRange"][1][:], self.state["viewRange"][2][:]]

    def setRange(self, xRange=None, yRange=None, zRange=None, padding=None, update=True, disableAutoRange=True):
        changes = {}   # axes
        setRequested = [False]*3
        if not xRange is None:
            changes[0] = xRange
            setRequested[0] = True
        if not yRange is None:
            changes[1] = yRange
            setRequested[1] = True
        if not zRange is None:
            changes[2] = zRange
            setRequested[2] = True
        if len(changes) == 0:
            raise Exception("Must specify at least one of rect, xRange, yRange or zRange")
                
        changed = [False]*3
        
   
        # Disable auto-range for each axis that was requested to be set
        if disableAutoRange:
            xOff = False if setRequested[0] else None
            yOff = False if setRequested[1] else None
            zOff = False if setRequested[2] else None

            self.enableAutoRange(x=xOff, y=yOff, z=zOff)
            changed.append(True)


        for ax, range in changes.items():
            mn = min(range)
            mx = max(range)

            preserve = False
            if mn == mx:
                preserve = True
                dy = self.state["viewRange"][ax][1] - self.state["viewRange"][ax][0]
                if dy == 0:
                    dy = 1
                mn -= dy*0.5
                mx += dy*0.5
        
            if not preserve:
                if padding is None:
                    xpad = self.state["defaultPadding"][ax]
                else:
                    xpad = padding
                p = (mx-mn) * xpad
                mn -= p
                mx += p

            if self.state['targetRange'][ax] != [mn, mx]:
                self.state['targetRange'][ax] = [mn, mx]
                changed[ax] = True

        lockX, lockY, lockZ = setRequested
        if lockX and lockY and lockZ:
            lockX = False
            lockY = False
            lockZ = False

        self.updateViewRange(lockX, lockY, lockZ)
        if any(changed):
            if changed[0] and not self.state['autoRange'][0]:
                self._autoRangeNeedsUpdate = True
            elif changed[1] and not self.state['autoRange'][1]:
                self._autoRangeNeedsUpdate = True
            elif changed[2] and not self.state['autoRange'][2]:
                self._autoRangeNeedsUpdate = True
                
            self.sigStateChanged.emit(self)
            
        
    def updateAutoRange(self):
        if self._updatingRange:
            return
        self._updatingRange = True
        try:
            if not any(self.state['autoRange']):
                return

            targetRect = self.viewRange()
            fractionVisible = self.state['autoRange'][:]
            for i in [0,1,2]:
                if type(fractionVisible[i]) is bool:
                    fractionVisible[i] = 1.0

            childRange = None
            kwargs = {}
            orthoRange = [None, None, None]
            for ax in [0, 1, 2]:
                if self.state['autoVisibleOnly'][ax]:
                    orthoRange[ax] = [targetRect[x] for x in range(3) if x != ax]    

            childRange = self.childrenBounds(frac=fractionVisible, orthoRange=orthoRange)
            
            try:
                pass
            except:
                pass
            
            for ax, s in enumerate("xyz"):
                if not self.state['autoRange'][ax]:
                    continue
                xr = childRange[ax]
                if xr is not None:
                    wp = (xr[1] - xr[0]) * self.state["defaultPadding"][ax]
                    childRange[ax][0] -= wp
                    childRange[ax][1] += wp
                    targetRect[ax] = childRange[ax]
                    kwargs[f"{s}Range"] = targetRect[ax]

            if len(kwargs) == 0:
                return
            kwargs['padding'] = 0.0
            kwargs['disableAutoRange'] = False
            self.setRange(**kwargs)
        finally:
            (f"finally")
            self._autoRangeNeedsUpdate = False
            self._updatingRange = False
            
    def setZRange(self, min, max, padding=None, update=True):
        """
        Set the visible Z range of the view to [*min*, *max*].
        The *padding* argument causes the range to be set larger by the fraction specified.
        (by default, this value is between the default padding and 0.1 depending on the size of the ViewBox)
        """
        self.setRange(zRange=[min, max], update=update, padding=padding)

    def setYRange(self, min, max, padding=None, update=True):
        """
        Set the visible Y range of the view to [*min*, *max*].
        The *padding* argument causes the range to be set larger by the fraction specified.
        (by default, this value is between the default padding and 0.1 depending on the size of the ViewBox)
        """
        self.setRange(yRange=[min, max], update=update, padding=padding)

    def setXRange(self, min, max, padding=None, update=True):
        """
        Set the visible X range of the view to [*min*, *max*].
        The *padding* argument causes the range to be set larger by the fraction specified.
        (by default, this value is between the default padding and 0.1 depending on the size of the ViewBox)
        """
        self.setRange(xRange=[min, max], update=update, padding=padding)


    def childrenBounds(self, frac=[1., 1., 1.], orthoRange=(None, None, None), items: List[GL3DGraphicsItemMixin]=None):
        """
        Method is similar to pyqtgraph.ViewBox, except this method does calculate the bounds for all items and axis in 1 call
        
        """
        if items is None:
            items = self.addedItems
        itemBounds = []
        
        for item in items:
            if not item.visible():
                continue
 
            if hasattr(item, 'dataBounds') and not item.dataBounds is None:
                if frac is None:
                    frac = (1.0, 1.0, 1.0)
                
                bounds_unchecked = item.dataBounds(frac, orthoRange)
                bounds = []
                bounds_arr = []
                for b_i in bounds_unchecked:
                    
                    if (b_i is None or (b_i[0] is None or b_i[1] is None)
                        or not math.isfinite(b_i[0])
                        or not math.isfinite(b_i[1])
                        ):
                        bounds.append([b_i, False])
                        bounds_arr.append(b_i)
                    else:
                        bounds.append([b_i, True])
                        bounds_arr.append(b_i)
                itemBounds.append(bounds)
            else:
                if item.flags() & item.GLGraphicsItemFlag.ItemHasNoContents:
                    continue
                bounds = self.mapFromItemToView(item, item.boundingRect()).boundingRect()
                itemBounds.append((bounds, True, True, 0))

        range = [None, None, None]
        for item_bound in itemBounds:
            for idx, (bounds, flag) in enumerate(item_bound):
                if flag:
                    if not range[idx] is None:
                        range[idx] = [min(bounds[0], range[idx][0]), max(bounds[1], range[idx][1])]
                    else:
                        range[idx] = [bounds[0], bounds[1]]
        return range


    def enableAutoRange(self, axis=None, enable=True, x=None, y=None, z=None):
        
        if not x is None or not y is None or not z is None:
            if not x is None:
                self.enableAutoRange(GL3DViewBox.XAxis, x)
            if not y is None:
                self.enableAutoRange(GL3DViewBox.YAxis, y)
            if not z is None:
                self.enableAutoRange(GL3DViewBox.ZAxis, z)
            return

        if enable is True:
            enable = 1.0

        if axis is None:
            axes = [0, 1, 2]
        
        elif axis == GL3DViewBox.XAxis or axis == 'x':
            axes = [0]
        elif axis == GL3DViewBox.YAxis or axis == 'y':
            axes = [1]
        elif axis == GL3DViewBox.ZAxis or axis == 'z':
            axes = [2]
        else:
            raise Exception('axis argument must be GLViewBox.XAxis, GLViewBox.YAxis, or GLViewBox.XYAxes.')
        for ax in axes:
            if self.state['autoRange'][ax] != enable:
                if enable is False and self._autoRangeNeedsUpdate:
                    self.updateAutoRange()

                self.state['autoRange'][ax] = enable
                self._autoRangeNeedsUpdate |= (enable is not False)
                self.update()
        self.sigStateChanged.emit(self)
        
    def changeParent(self):
        self._updateView()


    @QtCore.Slot()
    def prepareForPaint(self):
        #autoRangeEnabled = (self.state['autoRange'][0] is not False) or (self.state['autoRange'][1] is not False)
        # don't check whether auto range is enabled here--only check when setting dirty flag.
        if self._autoRangeNeedsUpdate: # and autoRangeEnabled:
            self.updateAutoRange()
        self.updateMatrix()
        
    def _updateView(self):
        ## called to see whether this item has a new view to connect to
        ## NOTE: This is called from GraphicsObject.itemChange or GraphicsWidget.itemChange.

        if not hasattr(self, '_connectedView'):
            # Happens when Python is shutting down.
            return

        ## It is possible this item has moved to a different GLViewBox or widget;
        ## clear out previously determined references to these.
        self.forgetViewBox()
        self.forgetViewWidget()
        
        ## check for this item's current viewbox or view widget
        view = self.getViewBox()
        #if view is None:
            ### "  no view"
            #return

        oldView = None
        if self._connectedView is not None:
            oldView = self._connectedView()
            
        if view is oldView:
            ## "  already have view", view
            return

        ## disconnect from previous view
        if oldView is not None:
            Device = 'Device' if hasattr(oldView, 'sigDeviceRangeChanged') else ''
            for signal, slot in [(f'sig{Device}RangeChanged', self.viewRangeChanged),
                                 (f'sig{Device}TransformChanged', self.viewTransformChanged)]:
                try:
                    getattr(oldView, signal).disconnect(slot)
                except (TypeError, AttributeError, RuntimeError):
                    # TypeError and RuntimeError are from pyqt and pyside, respectively
                    pass
            
            self._connectedView = None

        ## connect to new view
        if view is not None:
            view.sigRangeChanged.connect(self.viewRangeChanged)
            view.sigTransformChanged.connect(self.viewTransformChanged)
            self._connectedView = weakref.ref(view)
            self.viewRangeChanged()
            self.viewTransformChanged()
        
        self._replaceView(oldView)
        
        self.viewChanged(view, oldView)
        
    def viewChanged(self, view, oldView):
        """Called when this item's view has changed
        (ie, the item has been added to or removed from a GLViewBox)"""
        
    def _replaceView(self, oldView, item=None):
        if item is None:
            item = self
        for child in item.childItems():
            if isinstance(child, GL3DGraphicsItemMixin):
                if child.getViewBox() is oldView:
                    child._updateView()
                        #self._replaceView(oldView, child)
            else:
                self._replaceView(oldView, child)
        

    def forgetViewBox(self):
        self._viewBox = None

    def forgetViewWidget(self):
        self._viewWidget = None

    def mapFromItemToView(self, item, obj):
        """Maps *obj* from the local coordinate system of *item* to the view coordinates"""
        self.updateMatrix()
        return self.childGroup.mapFromItem(item, obj)


    def updateMatrix(self, changed=None):
        if not self._matrixNeedsUpdate:
            return

        vb_viewRange = self.viewRangeArr() 
        vb_worldRange = self.worldRangeArr() 

        range_view = vb_viewRange[:, 1] - vb_viewRange[:, 0]
        range_view[range_view == 0] = 1.0

        range_world = vb_worldRange[:, 1] - vb_worldRange[:, 0]
        range_world[range_world == 0] = 1.0

        scale = range_world / range_view
        offset = vb_worldRange[:, 0] - vb_viewRange[:, 0] * scale

        self.childGroup.resetTransform()
        tr = self.childGroup.transform()
        tr.translate(*offset)
        tr.scale(*scale)
        
        self.childGroup.setTransform(tr)
        self._matrixNeedsUpdate = False
        self.sigTransformChanged.emit(self)




    def viewWorldRange(self):
        vr0 = self.state["viewRange"][0]
        vr1 = self.state["viewRange"][1]
        vr2 = self.state["viewRange"][2]
        
        return QtCore.QRectF(vr0[0], vr1[0], vr2[0], vr0[1]-vr0[0], vr1[1]-vr1[0], vr2[1]- vr2[0])    
    
    def autoRangeEnabled(self):
        return self.state['autoRange'][:]

    def viewWorldRect(self):
        vr0 = self.state["viewRange"][0]
        vr1 = self.state["viewRange"][1]
        vr2 = self.state["viewRange"][2]
        
        return QtCore.QRectF(vr0[0], vr1[0], vr2[0], vr0[1]-vr0[0], vr1[1]-vr1[0], vr2[1]- vr2[0])
    
    def zoomWorld(self, ev, delta):
        for i in [1, 1/4]:
            rMultiplier = self._mouseZoomFactor**(i*delta)
            parent = self.parentWidget()
            distance = parent.opts['distance'] * rMultiplier
            opts = parent.opts.copy()
            opts["distance"] = distance
            opts["elevation"] = parent.opts["elevation"]
            opts["azimuth"] = parent.opts["azimuth"]

            if self._verifyOutsideBox(opts):
                parent.opts["distance"] = distance
                self.update()
            #    self.requestViewUpdate()
                zoomChanged=True
            else:
                zoomChanged=False
                            
            if zoomChanged:
                return 
        
    def zoomAxis(self, ev, delta):
        self.parentWidget().makeCurrent() 
        if self.mouse_state.constantAxis():
            pxCoords = ev.position().x(), ev.position().y()
            world_coords = GL3DViewBox.map_2D_coords_to_3D(self.parentWidget(), *pxCoords) 
            intersectedAxis, intersectedPoint = self.findIntersectingAxis(self.cameraPositionNumpy(), world_coords)
        else:
            wr = self.state["worldRange"]
            intersectedPoint = [0.5 * (wr[0][0] + wr[0][1]),
                                0.5 * (wr[1][0] + wr[1][1]),
                                0.5 * (wr[2][0] + wr[2][1]),
                                ]
            intersectedAxis = None

        s = 1.02 ** (delta * self.state['wheelScaleFactor'])
        
        s = [s if ax != intersectedAxis else None for ax in range(3)]

        intersectedPoint = self.childGroup.mapFromView(pg.Vector(*intersectedPoint))
        
        self._resetTarget()
        self.scaleBy(s=s, center=(intersectedPoint.x(), intersectedPoint.y(), intersectedPoint.z()))
        self.state["autoRange"] = 3 * [False]



    def scaleBy(self, **kwargs):
        self.scalebycounter+=1
        s, center, x, y, z = kwargs.get("s", None), kwargs.get("center", None), kwargs.get("x", None), kwargs.get("y", None), kwargs.get("z", None)

        if s is not None:
            x, y, z = s[0], s[1], s[2]

        affect = [x is not None, y is not None, z is not None]
        if not any(affect):
            return

        scaled_axes = [i for i in range(3) if affect[i]]
        if not scaled_axes:
            return

        # Get current view ranges
        view_ranges = self.viewRange()
        
        new_ranges = [None, None, None]

        for ax in scaled_axes:
            scale_factor = [x, y, z][ax]
            current_min, current_max = view_ranges[ax]
            width = current_max - current_min
            
            c = center[ax]
            new_width = width * scale_factor
            new_min = c - (c - current_min) * scale_factor
            new_max = c + (current_max - c) * scale_factor
            new_ranges[ax] = [new_min, new_max]
            print(f"c: {c}")

        set_kwargs = {}
        if new_ranges[0] is not None:
            set_kwargs['xRange'] = new_ranges[0]
        if new_ranges[1] is not None:
            set_kwargs['yRange'] = new_ranges[1]
        if new_ranges[2] is not None:
            set_kwargs['zRange'] = new_ranges[2]
        from pprint import pprint
        pprint(np.round(view_ranges, 3))
        pprint(set_kwargs)
        self.setRange(**set_kwargs, padding=0)
        self._matrixNeedsUpdate=True
        self.update()
        
    def mouseEnabled(self):
        return self.state['mouseEnabled'][:]
 
    def setDefaultPadding(self, padding=0.02):
        """
        Sets the fraction of the data range that is used to pad the view range in when auto-ranging.
        By default, this fraction is 0.02.
        """
        self.state['defaultPadding'] = padding

    def wheelEvent(self, ev: QtGui.QWheelEvent):
        delta = ev.angleDelta().x()
        if delta == 0:
            delta = ev.angleDelta().y()
        self.zoomAxis(ev, delta)
        self.parentWidget().update()
    
    def _emit_view_changed(self):
        self._last_update_time = QtCore.QTime.currentTime().msecsSinceStartOfDay()
        self._pending_update = False
        self.sigViewChange.emit(self)
        
    def requestViewUpdate(self):
        current_time = QtCore.QTime.currentTime().msecsSinceStartOfDay()
        elapsed = current_time - self._last_update_time
        
        if elapsed > self._update_interval:
            self._emit_view_changed()
        else:
            if not self._pending_update:
                self._pending_update = True
                QtCore.QTimer.singleShot(self._update_interval - elapsed,
                                         self._delayed_view_update)
                
    def _delayed_view_update(self):
        if self._pending_update:
            self._emit_view_changed()
   
        
    def _ProcessViewDrag(self, dP, center):
        if (np.abs(dP) > np.amin(self.state["worldSize"])).any():
            return 
        mask = [1.,]*3
        mask[self.mouse_state.intersectedAxis] = 0.
        mask = np.array(mask)
        
        s = ((mask * 0.2) + 1) ** dP   
        dx, dy, dz = s
        self._resetTarget()
        
        self.scaleBy(x=dx, y=dy, z=dz, center=center)
        self._matrixNeedsUpdate=True

    def mousePressEvent(self, ev):
        self.mouse_state.mouseToggleEvent(ev)
    
    def mouseReleaseEvent(self, ev):
        self.mouse_state.mouseToggleEvent(ev)

    def mouseMoveEvent(self, ev):
        if self.userInteracting():
            if bin(ev.buttons().value).count("1") == 1:
                mode = self.mouse_state.getModType(ev.buttons())
                if mode == "scale":
                    pos = ev.pos().x(), ev.pos().y()
                    camera_pos = self.cameraPositionNumpy()
                    self.parentWidget().makeCurrent() 
                    world_coords = GL3DViewBox.map_2D_coords_to_3D(self.parentWidget(), *pos)   
                    intersectedAxis, intersectedPoint = self.findIntersectingAxis(camera_pos, world_coords)
                    if not intersectedAxis is None:
                        intersectedPoint = np.array([point if idx != intersectedAxis else 0 for idx, point in enumerate(intersectedPoint)])
                        if not self.mouse_state.validIntersectedPoint():
                            self.mouse_state.setIntersectPoint(intersectedAxis, intersectedPoint)
                        else:
                            self.mouse_state.updateIntersectPoint(intersectedPoint)
                            delta = -1 * self.mouse_state.intersectedPoint.delta()
                            p_0 = self.mouse_state.intersectedPoint.initPos
                            self._ProcessViewDrag(delta, p_0)
                            
                elif mode == "pan":
                    pos = ev.pos().x(), ev.pos().y()
                    camera_pos = self.cameraPositionNumpy()
                    self.parentWidget().makeCurrent() 
                    world_coords = GL3DViewBox.map_2D_coords_to_3D(self.parentWidget(), *pos)   
                    intersectedAxis, intersectedPoint = self.findIntersectingAxis(camera_pos, world_coords)
                    if not intersectedAxis is None:
                        intersectedPoint = np.array([point if idx != intersectedAxis else 0 for idx, point in enumerate(intersectedPoint)])
                        if not self.mouse_state.validIntersectedPoint():
                            self.mouse_state.setIntersectPoint(intersectedAxis, intersectedPoint)
                        else:
                            self.mouse_state.updateIntersectPoint(intersectedPoint)
                            delta = -1 * self.mouse_state.intersectedPoint.delta()
                            self.translateBy(**{ax : delta[i] for i, ax in enumerate("xyz") if i != intersectedAxis} )
                self.state["autoRange"] = 3 * [False]
                return 

    
    def translateBy(self, t=None, x=None, y=None, z=None):
        """
        Translate the view by *t*, which may be a Point or tuple (x, y).

        Alternately, x or y may be specified independently, leaving the other
        axis unchanged (note that using a translation of 0 may still cause
        small changes due to floating-point error).
        """
        vr = self.targetRange()
        for i, ax in enumerate((x, y, z)):
            if not ax is None:
                vr[i] += ax
        kwargs = {}
        
        if not x is None:
            kwargs["xRange"] = vr[0]
        if not y is None:
            kwargs["yRange"] = vr[1]
        if not z is None:
            kwargs["zRange"] = vr[2]
        self.setRange(**kwargs, padding=0)

    @classmethod
    def zoom_2d(cls, point, drag, viewRange, base=1.1):
        axis_mins = np.array([vr[0] for vr in viewRange])
        axis_maxs = np.array([vr[1] for vr in viewRange])

        scales = base ** -drag
        
        new_axis_mins = scales * axis_mins + (1 - scales) * point
        new_axis_maxs = scales * axis_maxs + (1 - scales) * point
        return [[new_min, new_max] for new_min, new_max in zip(new_axis_mins, new_axis_maxs)]


    def _verifyOutsideBox(self, opts):
        if any([self.state["worldRange"][idx][1] <= pos for idx, pos in enumerate(self.parentWidget().cameraPosition(opts))]):
            return True
        else:
            return False
        
    def _checkCartesianBounds(self, coords):
        return [self.state["worldRange"][i][0] <= coords[i] < self.state["worldRange"][i][1] for i in range(3)]

    def keyPressEvent(self, ev):
        ev.accept()
        self.mouse_state.KeyToggleEvent(ev)


    def keyReleaseEvent(self, ev):
        self.mouse_state.KeyToggleEvent(ev)
        
            
    def focusOutEvent(self, event):...
        #print("OpenGLWidget lost focus!")

    def cameraPositionNumpy(self):
        p = self.parentWidget().cameraPosition()
        self._camera_position[:] = p.x(), p.y(), p.z()
        return self._camera_position
    
    
    def targetRect(self, ax):
        """
        Return the region which has been requested to be visible.
        (this is not necessarily the same as the region that is *actually* visible--
        resizing and aspect ratio constraints can cause targetRect() and viewRect() to differ)
        """
        try:
            tr0, tr1 = [self.state['targetRange'][i] for i in range(3) if i != ax]

            return QtCore.QRectF(tr0[0], tr1[0], tr0[1]-tr0[0], tr1[1] - tr1[0])
        except:
            raise

    @classmethod
    def line_rectangle_intersection(cls, P1, P2, rect_corners):
        P1 = np.array(P1)
        P2 = np.array(P2)
        rect_corners = [np.array(c) for c in rect_corners]
        
        v1 = rect_corners[1] - rect_corners[0]
        v2 = rect_corners[2] - rect_corners[0]
        normal = np.cross(v1, v2)
        normal = normal / np.linalg.norm(normal)

        # Plane equation: (P - P0) · normal = 0
        P0 = rect_corners[0]
        
        # Line direction
        line_dir = P2 - P1
        line_dir = line_dir / np.linalg.norm(line_dir)
        
        # Check if line is parallel to the plane (no intersection)
        denom = np.dot(normal, line_dir)
        if np.abs(denom) < 1e-6:
            return None  # Line is parallel to the plane
        
        # Compute intersection point on the plane
        t = np.dot(normal, P0 - P1) / denom
        intersection = P1 + t * line_dir
        
        # Check if intersection point is within the rectangle
        # Project the point onto the rectangle's plane and check barycentric coordinates
        u = np.dot(intersection - P0, v1) / np.dot(v1, v1)
        v = np.dot(intersection - P0, v2) / np.dot(v2, v2)
        
        if 0 <= u <= 1 and 0 <= v <= 1:
            return intersection
        else:
            return None

    @classmethod
    def map_2D_coords_to_3D(cls, widget: "GL3DViewWidget", x: float, y: float):
        widget_width = widget.width()
        widget_height = widget.height()
        device_pixel_ratio = widget.window().screen().devicePixelRatio()
    
        ndc_x = x / widget_width
        ndc_y = y / widget_height

        viewport = GL.glGetIntegerv(GL.GL_VIEWPORT)
        
        _, _, viewport_width, viewport_height = viewport

        mouse_x_physical = ndc_x * viewport_width
        mouse_y_physical = ndc_y * viewport_height
        mouse_y_physical = viewport_height - mouse_y_physical 
        
        
        gl_px_width, gl_px_height = 1, 1
        
        depth = GL.glReadPixels(int(mouse_x_physical), int(mouse_y_physical), gl_px_width, gl_px_height, GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT)[0][0]
        modelview = np.array(widget.viewMatrix().data()).reshape(4, 4)
        projection = np.array(widget.projectionMatrix(viewport, (0, 0, device_pixel_ratio * widget_width, device_pixel_ratio* widget_height)).data()).reshape(4, 4)
        depth = 0.
        world_x, world_y, world_z = GLU.gluUnProject(mouse_x_physical, mouse_y_physical, depth, modelview, projection, viewport)
        return world_x, world_y, world_z

    @classmethod
    def map_3D_coords_to_2D(cls, widget, world_x, world_y, world_z):
        device_pixel_ratio = widget.window().screen().devicePixelRatio()
        widget_height = widget.height()
        widget_width = widget.width()
        
        viewport = GL.glGetIntegerv(GL.GL_VIEWPORT)
        _, _, viewport_width, viewport_height = viewport
        
        modelview = np.array(widget.viewMatrix().data()).reshape(4, 4)
        projection = np.array(widget.projectionMatrix(viewport,
                                                     (0, 0, device_pixel_ratio * widget_width, device_pixel_ratio * widget_height)).data()).reshape(4, 4)
        
        px_x, px_y, px_z = GLU.gluProject(world_x, world_y, world_z, modelview, projection, viewport)
        px_y = viewport_height - px_y
        return px_x, px_y, px_z
    
    def FacingVerticesPx(self):
        closest_faces = self.closestFaces()
        view_widget = self.parentWidget()

        ax_boundingRectPx = []
        for ax, face in enumerate(closest_faces):
            vertexCoords = self.state["vertexCoordinates"][ax][face]
            screen_coords = []
            for coord in vertexCoords:
                x_px, y_px, z_px = self.map_3D_coords_to_2D(view_widget, *coord)
                screen_coords.append([x_px, y_px])
                
            ax_boundingRectPx.append(screen_coords)

        ax_boundingRectPx = np.array(ax_boundingRectPx).round(4).tolist()
        return ax_boundingRectPx

    def extendPoint(self, p1, p2):
        p2_new = [0, 0, 0]
        dx = p2[0] - p1[0]
        dy = p2[1] - p1[1]
        dz = p2[2] - p1[2]
        d = (dx**2 + dy**2 + dz**2)**0.5    
        dNew = 2 * self.parentWidget().opts["distance"]
        p2_new[0] = p1[0] + dx / d * dNew
        p2_new[1] = p1[1] + dy / d * dNew
        p2_new[2] = p1[2] + dz / d * dNew
        return p1, p2
    
    def findIntersectingAxis(self, p1, p2):
        p1, p2 = self.extendPoint(p1, p2)
        closest_faces = self.closestFaces()
        vertexCoordinates = self.state["vertexCoordinates"]
        intersectionAxis = None
        intersectingPoint = None
        for ax in range(3):
            face_side = closest_faces[ax]
            rect = vertexCoordinates[ax][face_side]
            intersection = GL3DViewBox.line_rectangle_intersection(p1, p2, rect)
            if not intersection is None:
                intersectionAxis=ax
                intersectingPoint=intersection
                break
        
        return intersectionAxis, intersectingPoint

    def screen_coords(self, world_coordinates):
        if len(world_coordinates) == 3:
            world_coordinates = np.append(world_coordinates, 1)
        mvp = self.parentWidget().currentProjection() * self.parentWidget().currentModelView()  
        return np.array(mvp.data()).reshape(4, 4) @ world_coordinates
            
    def closestFaces(self):
        cameraPosition = self.parentWidget().cameraPosition()
        worldRange = self.state["worldRange"]
        pos = cameraPosition.x(), cameraPosition.y(), cameraPosition.z()

        faces = []
        for ax in range(3):
            ax_l, ax_p = worldRange[ax]
            p = pos[ax]
            ax_face = 0 if (p - ax_l) < (ax_p - p) else 1
            faces.append(ax_face)
        return faces

    def get_cube_face2(self, worldRange):
        axis_px_rect = []
        
        for ax in range(3):
            
            inner_rect = []
            for ax_lim in worldRange[ax]:
                wr_i = np.delete(worldRange, ax, axis=0)
                point = [np.insert([wr_i[i][j], wr_i[j][i]], ax, ax_lim) for j in range(2) for i in range(2)]
                inner_rect.append(point)
            axis_px_rect.append(inner_rect)
                
        return axis_px_rect

    def _childPaint(self):...
    
    def _paintHelper(self):... 

    def get_cube_face(self, worldRange):
        axis_px_rect = []
        
        for ax in range(3):
            # Get the other two axes (e.g., if ax=0 (x), then other_axes=[1,2] (y,z))
            other_axes = [i for i in range(3) if i != ax]
            ax1, ax2 = other_axes
            
            inner_rect = []
            for ax_lim in worldRange[ax]:  # min and max of current axis
                # Get min/max for the other two axes
                ax1_min, ax1_max = worldRange[ax1]
                ax2_min, ax2_max = worldRange[ax2]
                
                # Generate the 4 vertices of this face
                # (No longer using 'j' inside a comprehension)
                face_vertices = [
                    [ax_lim, ax1_min, ax2_min],  # Vertex 1
                    [ax_lim, ax1_max, ax2_min],  # Vertex 2
                    [ax_lim, ax1_max, ax2_max],  # Vertex 3
                    [ax_lim, ax1_min, ax2_max],  # Vertex 4
                ]
                
                # Reorder axes to match original structure (x,y,z)
                # (Since we hardcoded ax1, ax2, we need to restore the correct order)
                ordered_face = []
                for vertex in face_vertices:
                    # Reconstruct the vertex in x,y,z order
                    new_vertex = [0, 0, 0]
                    new_vertex[ax] = vertex[0]          # Fixed axis (ax_lim)
                    new_vertex[ax1] = vertex[1]         # First other axis
                    new_vertex[ax2] = vertex[2]         # Second other axis
                    ordered_face.append(new_vertex)
                
                inner_rect.append(ordered_face)
            
            axis_px_rect.append(inner_rect)
        
        return axis_px_rect
    
    def userInteracting(self): return self.mouse_state.keyInteracting()
    
    
    
def window_to_world_ray(view: "GL3DViewWidget", wx: float, wy: float):
    """Convert a window point (wx, wy) to a 3D ray (origin, direction) in world space."""
    # Get modelview and projection matrices
    modelview = view.viewMatrix()
    projection = view.projectionMatrix()

    # Compute the inverse of (projection * modelview)
    mvp = projection @ modelview
    inv_mvp = np.linalg.inv(mvp)

    # Convert window coordinates to normalized device coordinates (NDC)
    # Note: PyQtGraph uses Qt's coordinate system where Y increases downward.
    # So, we may need to flip Y if necessary.
    width, height = view.width(), view.height()
    ndc_x = (2.0 * wx) / width - 1.0
    ndc_y = 1.0 - (2.0 * wy) / height  # Flip Y to match OpenGL NDC

    # Near and far points in NDC
    near_ndc = np.array([ndc_x, ndc_y, -1.0, 1.0])  # z = -1 (near plane)
    far_ndc = np.array([ndc_x, ndc_y, 1.0, 1.0])    # z = 1 (far plane)

    # Transform to world coordinates
    near_world = inv_mvp @ near_ndc
    near_world /= near_world[3]  # Perspective divide

    far_world = inv_mvp @ far_ndc
    far_world /= far_world[3]

    # Ray origin and direction
    origin = near_world[:3]
    direction = far_world[:3] - origin
    direction /= np.linalg.norm(direction)  # Normalize

    return origin, direction
       

def ray_intersects_cube(origin, direction, cube_bounds):
    """Check if a ray intersects a cube defined by [[xmin, xmax], [ymin, ymax], [zmin, zmax]]."""
    tmin = -np.inf
    tmax = np.inf
    
    for i in range(3):  # Check x, y, z axes
        if np.abs(direction[i]) < 1e-6:  # Ray parallel to slab
            if origin[i] < cube_bounds[i][0] or origin[i] > cube_bounds[i][1]:
                return False
        else:
            t1 = (cube_bounds[i][0] - origin[i]) / direction[i]
            t2 = (cube_bounds[i][1] - origin[i]) / direction[i]
            tmin = max(tmin, min(t1, t2))
            tmax = min(tmax, max(t1, t2))
    
    return tmax >= tmin and tmax >= 0  # Intersection exists and is in front of the camera

def is_window_point_inside_cube(view: "GL3DViewWidget", wx: float, wy: float, cube_bounds):
    """Check if the window point (wx, wy) is inside the cube."""
    origin, direction = window_to_world_ray(view, wx, wy)
    return ray_intersects_cube(origin, direction, cube_bounds)


import sys
import numpy as np
from PySide6.QtWidgets import QApplication
from PySide6.QtCore import QTimer
import pyqtgraph.opengl as gl
from PySide6.QtGui import QVector3D, QMatrix4x4

class TransformDemo(gl.GLViewWidget):
    def __init__(self):
        super().__init__()
        self.setCameraPosition(distance=15)

        # Axis grid
        grid = gl.GLGridItem()
        grid.setSize(10, 10)
        grid.setSpacing(1, 1)
        self.addItem(grid)

        # Create an "L" shaped object made of 2 points
        self.l_shape = np.array([
            [2, 0, 0],  # horizontal arm
            [2, 1, 0],  # vertical arm
        ])

        # Compute centroid of L shape
        self.centroid_local = np.mean(self.l_shape, axis=0)

        # Scatter 1 - Local (centered at origin)
        pos_local = self.l_shape - self.centroid_local  # shift points so centroid is at origin
        self.scatter_local = gl.GLScatterPlotItem(pos=pos_local,
                                                  color=(1, 0, 0, 1))
        self.scatter_local.setGLOptions('opaque')
        self.addItem(self.scatter_local)

        # Position the local scatter at centroid location
        self.scatter_local.translate(*self.centroid_local)

        # Scatter 2 - Global (shifted in space)
        pos_global = self.l_shape + np.array([[5, 0, 0]])
        self.scatter_global = gl.GLScatterPlotItem(pos=pos_global,
                                                   color=(0, 1, 0, 1))
        self.scatter_global.setGLOptions('opaque')
        self.addItem(self.scatter_global)

        # Timer
        self.timer = QTimer()
        self.timer.timeout.connect(self.apply_rotation)
        self.timer.start(1000)

    def apply_rotation(self):
        # ----- RED: Local Rotation around its own centroid -----
        # Step 1: move to origin (inverse translate)
        """
        self.scatter_local.translate(-self.centroid_local[0],
                                    -self.centroid_local[1],
                                    -self.centroid_local[2])
        
        """
        self.scatter_local.rotate(45, 0, 0, 1, local=True)

        # Step 3: move back to centroid
        """
        self.scatter_local.translate(self.centroid_local[0],
                                    self.centroid_local[1],
                                    self.centroid_local[2])
        """
        # ----- GREEN: Global Rotation around scene origin -----
        self.scatter_global.rotate(45, 0, 0, 1, local=False)

if __name__ == '__main__':
    app = QApplication(sys.argv)
    w = TransformDemo()
    w.show()
    sys.exit(app.exec())


from PySide6 import QtWidgets, QtCore, QtGui
import sys
import numpy as np

class OptionMetricCellItem(QtWidgets.QTableWidgetItem):
    def __init__(self, text=""):
        super().__init__(text)
        self.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
        self.setBackground(QtGui.QColor("black"))
        self.setForeground(QtGui.QColor("white"))
        self.setTextAlignment(QtCore.Qt.AlignRight)

class OptionMetrics(QtWidgets.QTableWidget):
    def __init__(self, instrument_name, ins_ob, price_types, parent=None):
        super().__init__(parent)
        self.instrument_name = instrument_name
        self.price_types = price_types
        self.popped=False
        
        
        self.setRowCount(2)
        self.setColumnCount(2)
        
        self.setHorizontalHeaderLabels(price_types)
        self.setVerticalHeaderLabels([instrument_name, "Value"])
        counter=0
        for i in range(2):
            for j in range(2):
                value = ins_ob.values[counter]
                counter+=1
                self.setItem(i, j, OptionMetricCellItem(f"{value:.2f}"))
    def set_on_view(self, flag):
        self._on_view=flag
        
    def update(self, item):
        
        if self._on_view:
            counter=0
            for i in range(2):
                for j in range(2):
                    value = item.values[counter]
                    counter+=1
                    cell_item = self.item(i,j)
                    cell_item.setText(f"{value:.2f}")

    
    
class OptionMetricsWithTabs(QtWidgets.QWidget):
    def __init__(self, instruments, instrument_container, parent=None):
        super().__init__(parent)
        self.name_tab_idx_map={}
        self.instruments=instruments
        
        self.setup_ui(instrument_container)
        
    def _popout_table(self, pos, button):
        instrument_name = button.text()
        idx = self.name_tab_idx_map[instrument_name]
        table = self.stack.itemAt(idx).widget()
        table.set_on_view(True)
        
        app = QtWidgets.QApplication.instance()
        table.popped=True
        
        self.button_layout.removeWidget(button)
        button.deleteLater()  
        del self.name_tab_idx_map[instrument_name]
        self.button_group.removeButton(button)
        
            
        for name, i in list(self.name_tab_idx_map.items()):
            if idx < i:
                button = self.button_group.buttons()[i-1]
                new_i = self.name_tab_idx_map[name] - 1
                self.button_group.setId(button, new_i)
                self.name_tab_idx_map[name]=new_i
 
        self.stack.removeWidget(table)
        table.show()
        app.open_sub_window(pos, table)

        for col in range(table.columnCount()):
            for row in range(table.rowCount()):
                item = table.item(row, col)
                print(f"\n{item.text()}")
        
    def _open_button_context(self, pos):
        button = self.sender()
        
        menu = QtWidgets.QMenu(self)
        
        print_action = menu.addAction("Open in New Window")
        print_action.triggered.connect(lambda: self._popout_table([pos.x(), pos.y()], button))
        menu.exec(button.mapToGlobal(pos))

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self.adjust_button_layout()
        
    def adjust_button_layout(self):
        if not hasattr(self, 'button_group') or not self.button_group.buttons():
            return
        
        table = self.stack.currentWidget()
        for btn in self.button_group.buttons():
            btn.hide()
        table.adjustSize()
        available_width = table.sizeHint().width()
        
        
        test_button = self.button_group.buttons()[0]
        button_width = test_button.sizeHint().width()

        print(f"button_width: {button_width}")
        print(f"available_width: {available_width}")
        print(f"test_button.size().width(): {test_button.size().width()}")
        spacing = self.button_layout.spacing()
        
        total_button_width = button_width + spacing
        for btn in self.button_group.buttons():
            btn.show()

        
        
        max_buttons = max(1, int(available_width / total_button_width))
        
        current_buttons_per_row = self.button_layout.count()
        
        if max_buttons != current_buttons_per_row:
            self.rearrange_buttons(max_buttons)
            
    def rearrange_buttons(self, buttons_per_row):
        while self.button_layout.count():
            item = self.button_layout.takeAt(0)
            if item.widget():
                item.widget().setParent(None)
        
        grid_layout = QtWidgets.QGridLayout()
        grid_layout.setSpacing(self.button_layout.spacing())
        
        buttons = self.button_group.buttons()
        for i, button in enumerate(buttons):
            row = i // buttons_per_row
            col = i % buttons_per_row
            grid_layout.addWidget(button, row, col)
        
        self.main_layout.removeItem(self.button_layout)
        self.button_layout = grid_layout
        self.main_layout.insertLayout(0, self.button_layout)
        self.button_layout.setSpacing(0)

        
    def setup_ui(self, instrument_container):
        self.main_layout=QtWidgets.QVBoxLayout(self)
        self.setLayout(self.main_layout)
        
        self.button_group = QtWidgets.QButtonGroup()
        self.button_layout = QtWidgets.QHBoxLayout()
        
        for i, (name, _) in enumerate(self.instruments.items()):
            self.name_tab_idx_map[name]=i
            btn = QtWidgets.QPushButton(name)
            btn.setCheckable(True)
            btn.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)  
            btn.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
            btn.customContextMenuRequested.connect(self._open_button_context)
            self.button_group.addButton(btn, i)
            self.button_layout.addWidget(btn)
        
        self.stack = QtWidgets.QStackedLayout()
        for ins_ob in instrument_container:
            name = ins_ob.instrument_name
            data = self.instruments[name]

            table = OptionMetrics(name, ins_ob, data["price_types"])
            print(f"table: {table}")
            ins_ob.add_callback(table.update)
            table.set_on_view(False)
            instrument_container
            self.stack.addWidget(table)
            self.stack.currentWidget().width()
    
        self.button_group.idClicked.connect(self.stack.setCurrentIndex)
        
        
        
        self.main_layout.addLayout(self.stack)
        
        
        if self.button_group.buttons():
            self.button_group.buttons()[0].setChecked(True)
        
        self.stack.itemAt(0).widget().set_on_view(True)
        self.subwindow_width = self.stack.itemAt(0).widget().width()
        self.main_layout.insertLayout(0, self.button_layout)


window_container=[]

class Application(QtWidgets.QApplication):
    create_window_signal = QtCore.Signal(dict, str, tuple)
    window_titles_signal = QtCore.Signal(list)
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance
    
    @classmethod
    def get_instance(cls):
        return cls._instance
    
    def __init__(self,
                 argv
                 ):
        super().__init__(argv)
        
    def open_main_window(self):
        for i in range(1, 9999):
            window_title = f"Main Window {i}"
        win = QtWidgets.QMainWindow()

        window_container.append(win)
        win.showMaximized()
    
    def open_sub_window(self, geometry, widget):
        for i in range(1, 9999):
            window_title = f"Sub Window {i}"
        win = SubWindow(geometry, widget)
        window_container.append(win)

class SubWindow(QtWidgets.QMainWindow):
    def __init__(self, geometry, widget):
        super().__init__()
        self.widget_main = widget
        self.setCentralWidget(self.widget_main)
        
        self.layout_main = QtWidgets.QVBoxLayout(self.widget_main)
        self.setLayout(self.layout_main)
        frame = self.frameGeometry()
        frame_size = frame.size()
        widget_size = self.widget_main.sizeHint()
        
        width_diff = frame_size.width() - self.centralWidget().sizeHint().width()
        height_diff = frame_size.height() - self.centralWidget().sizeHint().height()
        
        self.resize(widget_size.width() + width_diff, 
                    widget_size.height() + height_diff)
        
        self.move(geometry[0], geometry[1])
        self.show()       
        
        

            
        
class Updater(QtCore.QThread):
    def __init__(self, instruments):
        self.instruments=instruments
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update)
        self.timer.start(1000)

    def update(self):
        for instrument in self.instruments:
            instrument.update()
             

class InsObj:
    def __init__(self, instrument_name, S_0):
        self.instrument_name=instrument_name
        self.S_0=S_0
        self.values = self.S_0 + np.random.normal(0, 0.3, size=4)
        self.callbacks=[]
        
    def add_callback(self, callback):
        self.callbacks.append(callback)
    
    def update(self):
        self.values = self.values + np.random.normal(0, .3, size=4)
        for c in self.callbacks:
            c(self)
        
if __name__ == "__main__":
    app = Application(sys.argv)
    
    window = QtWidgets.QMainWindow()
    instruments = {
            "BTC-25JUN-70000-C": {
                "object": None,  
                "price_types": ["Bid", "Ask"]
            },
            "BTC-25JUN-75000-C": {
                "object": None,
                "price_types": ["Mid", "Mark"]
            },
            "BTC-25JUN-80000-C": {
                "object": None,
                "price_types": ["Bid", "Ask"]
            }
        }

    
    S_0 = 10
    K = [70000 + i*5000 for i in range(1,8)]
    instrument_container=[]
    instruments={}
    for i in range(1, 8):
        k = K[i-1]
        instruments[f"BTC-25JUN-{k}-C"] = {"object": None,
                                            "price_types": ["Bid", "Ask"]}
        S_0 = S_0 + 10 
        
        o = InsObj(f"BTC-25JUN-{k}-C", S_0)
        instrument_container.append(o)
    
    workers = Updater(instrument_container)
    
    central_widget = OptionMetricsWithTabs(instruments, instrument_container)
    window.setCentralWidget(central_widget)
    window.resize(400, 300)
    window.show()
    
    sys.exit(app.exec())


from __future__ import annotations
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from pyvol_terminal.windows import Window


from PySide6 import QtWidgets, QtCore
from . import utils



class ContextPushButton(QtWidgets.QPushButton):
    def __init__(self, label, parent=None):
        super().__init__(label, parent)
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(lambda pos: parent._show_button_menu(self, pos))

    def mousePressEvent(self, ev):
        if ev.button() == QtCore.Qt.RightButton:
            ev.accept()
            self.customContextMenuRequested.emit(ev.pos())
        else:
            super().mousePressEvent(ev)


class WindowSettings(QtWidgets.QWidget):
    def __init__(self,
                 new_window_slots: List[Callable],
                 interface_name_slot_map: Dict[str, Callable]):
        super().__init__()
        self.new_window_slots=new_window_slots
        self.interface_name_slot_map=interface_name_slot_map
        self.idx_to_interface_name = {}
        self._interface_to_idx={}
        
        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0 ,0)

        self.button_group = QtWidgets.QButtonGroup()
        btn=QtWidgets.QPushButton("Open New Window")
        btn.setCheckable(False)
        self.button_group.addButton(btn, 0)
        for slot in self.new_window_slots:
            btn.clicked.connect(slot)
        self.layout().addWidget(btn)
            
        self._create_interface_buttons()
        self.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        
    def _create_interface_buttons(self):        
        self.button_group.setExclusive(True)
        
        for idx, (name, slots) in enumerate(self.interface_name_slot_map.items(), start=1):
            self._interface_to_idx[name]=idx
            btn=QtWidgets.QPushButton(name)
            btn.setStyleSheet(utils.get_settings_stylesheets("QPushButton"))
            btn.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            btn.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
            
            btn.setCheckable(True)
            btn.setChecked(False)            
            
            self.button_group.addButton(btn, id=idx) 
            self.layout().addWidget(btn)
            
            for slot in slots:
                self.button_group.idToggled.connect(slot)
            
    def toggle_price_type(self,
                          idx: int=None,
                          name: str=None
                          ):
        if not idx is None:
            self.button_group.button(idx).toggle()
        else:
            if not name is None:
                print(self._interface_to_idx[name])
                print(self.button_group.button(self._interface_to_idx[name]))
                self.button_group.button(self._interface_to_idx[name]).toggle()
            
    @QtCore.Slot(QtWidgets.QPushButton)
    def left_click(self, button):
        btn_id = self.button_group.id(button)
        print(f"btn_id: {btn_id}")
        if btn_id == self.prev_btn_idx:
            return 
        else:
            self.prev_btn_idx = btn_id
        
        interface_name = self.idx_to_interface_name[btn_id]
        print(f"interface_name: {interface_name}")
        for callback in self.interface_name_callbacks_dict[interface_name]:
            callback(interface_name)
    
             
    def _popout_action(self, interface_name, btn_id):   
        interface = self.window_widget.interfaces[interface_name]
        if btn_id != self.prev_btn_idx:
            self.window_widget.price_process_worker.update_interface_signal.connect(interface.update_interface)
                 
        self.open_new_window(interface_name, interface, self.configs, self._collapse_action)
        btn = self.button_group.button(btn_id)
        self.button_group.removeButton(btn)
        self.layout().removeWidget(btn)
        self.buttons.remove(btn)
        btn.deleteLater()      
        other_interface_btn = self.buttons[0] 
        idx = self.button_group.id(other_interface_btn)
        other_interface = self.window_widget.interfaces[self.idx_to_interface_name[idx]]
        self.window_widget.price_process_worker.update_interface_signal.connect(other_interface.update_interface)
        self.left_click(other_interface_btn)
        
         
    
    def _collapse_action(self, interface_name, sub_window):
        print(f"\n_collapse_action")
        print(f"{self.main_window_flag}")
        print(f"self.idx_to_interface_name: {self.idx_to_interface_name}")
        self.window_widget.price_process_worker.update_interface_signal.disconnect(sub_window.interface.update_interface)
        sub_window.close()
        del sub_window
        btn = ContextPushButton(interface_name, parent=self)
        btn.setCheckable(True)
        btn.setStyleSheet(utils.get_settings_stylesheets("QPushButton"))
        btn.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.buttons.append(btn)
        self.layout().addWidget(btn)
        btn_id = self._interface_to_idx[interface_name]
        self.button_group.addButton(btn, id=btn_id) 
        btn.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
    
        
    def _show_button_menu(self, button, pos):
        menu = QtWidgets.QMenu(self)
        btn_id = self.button_group.id(button)
        iface = self.idx_to_interface_name[btn_id]
        menu.blockSignals(True)
        if self.main_window_flag:        
            menu.addAction("Pop-out Window", lambda: self._popout_action(iface, btn_id))
        else:
            menu.addAction("Collapse Window", lambda: self.main_window_collapse_func(iface, self.window_widget))
        menu.blockSignals(False)
        menu.exec(button.mapToGlobal(pos))
        



class CenteredPopupComboBox(QtWidgets.QComboBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def showPopup(self):
        super().showPopup()
        
        popup = self.view()
        current_index = self.currentIndex()

        if current_index >= 0:
            rect = popup.visualRect(popup.model().index(current_index, 0))
            scroll_position = rect.top() - (popup.viewport().height() - rect.height()) // 2
            popup.verticalScrollBar().setValue(scroll_position)

    def wheelEvent(self, event):
        if self.view().isVisible():
            super().wheelEvent(event)



class CustomWarnings(Warning):
    def __init__(self, message, code=None):
        super().__init__(message)
        self.code = code  

    def __str__(self):
        base_msg = super().__str__()
        return f"[Code {self.code}] {base_msg}" if self.code else base_msg

class InsufficientDataWarning(CustomWarnings):
    def __init__(self, message, code=100):
        super().__init__(message, code)

class InterpolationFitWarning(CustomWarnings):
    def __init__(self, message, code=200):
        super().__init__(message, code)

class InterpolationEvalWarning(CustomWarnings):
    def __init__(self, message, code=300):
        super().__init__(message, code)


from __future__ import annotations 
from pydantic.dataclasses import dataclass, Field
from dataclasses import InitVar
from pyvol_terminal.axis import axis_utils as axis_utils
import numpy as np
from pyvol_terminal import utils
from typing import Any, Dict, List, Optional
from pyvol_terminal import exceptions
import warnings
from abc import ABC, abstractmethod


def custom_showwarning(message, category, filename, lineno, file=None, line=None):
    print(f"UserWarning: {message} ({filename}, line {lineno})")
warnings.showwarning = custom_showwarning



class Config:
    arbitrary_types_allowed=True
    slots=True

        
class Config2:
    arbitrary_types_allowed = True
    frozen: True

    
class Config3:
    arbitrary_types_allowed=True
    kw_only=True
    slots=True


@dataclass(slots=True, config=Config2)
class BaseDomain:
    strike: np.ndarray
    expiry: np.ndarray
    x_metric: str
    y_metric: str
    

@dataclass(config=Config)
class SquaredDomain:
    base_domain: BaseDomain
    x: np.ndarray = Field(init=False, default_factory=lambda: np.array)
    y: np.ndarray = Field(init=False, default_factory=lambda: np.array)
    x_vect: np.ndarray = Field(init=False, default_factory=lambda: np.array)
    y_vect: np.ndarray = Field(init=False, default_factory=lambda: np.array)

    def __post_init__(self):
        self.x_vect = np.unique(self.base_domain.strike)
        self.y_vect = np.unique(self.base_domain.expiry)
        self.x, self.y = np.meshgrid(self.x_vect, self.y_vect, indexing="xy")

        
@dataclass(slots=True, config=Config)
class Domain:
    base_domain: BaseDomain
    x: np.ndarray = Field(init=False, default_factory=lambda: np.array)
    y: np.ndarray = Field(init=False, default_factory=lambda: np.array)
    x_vect: np.ndarray = Field(init=False, default_factory=lambda: np.array)
    y_vect: np.ndarray = Field(init=False, default_factory=lambda: np.array)
    x_mat: np.ndarray = Field(init=False, default_factory=lambda: np.array)
    y_mat: np.ndarray = Field(init=False, default_factory=lambda: np.array)
    z_mat: np.ndarray = Field(init=False, default_factory=lambda: np.array)

    xy: np.ndarray = Field(init=False, default_factory=lambda: np.array)
    idx_ij_mapping: Dict[int, tuple] = Field(init=False, default_factory=dict)
    z_mat: float = Field(init=False, default_factory=lambda: np.array)
    
    def __post_init__(self):
        self._create_domain(self.base_domain.strike.copy(), self.base_domain.expiry.copy())

    def _create_domain(self, x, y):
        self.x = x
        self.y = y
        self.xy = np.column_stack((self.x, self.y)) 

        self.x_vect = np.unique(self.base_domain.strike)
        self.y_vect = np.unique(self.base_domain.expiry)

        self.x_mat, self.y_mat = np.meshgrid(self.x_vect, self.y_vect, indexing="xy")
        self.z_mat = np.full(self.x_mat.shape, np.nan)
        
        self.idx_ij_mapping = {}
        
        for i in range(self.x_mat.shape[0]):
            for j in range(self.x_mat.shape[1]):
                match_idx = np.where((self.xy == [self.x_mat[i, j], self.y_mat[i, j]]).all(axis=1))[0]
                for idx in match_idx:
                    self.idx_ij_mapping[idx] = (i, j)

    def update_data(self, z, idx_map):
        for idx, ivol in zip(idx_map, z):
            if idx in self.idx_ij_mapping:
                i, j = self.idx_ij_mapping[idx]
                self.z_mat[i,j] = ivol 


@dataclass(config=Config)
class BaseData:
    n_instruments: int
    name_index_map: Dict[str, int]
    price: np.ndarray
    
        
@dataclass(config=Config)
class RawFutures(BaseData):
    OTM: np.ndarray


@dataclass(config=Config)
class Greeks:
    delta: np.ndarray
    delta_mag: np.ndarray
    gamma: np.ndarray
    vega: np.ndarray
    theta: np.ndarray
    rho: np.ndarray

@dataclass(config=Config)
class Moneyness:
    moneyness: np.ndarray
    log_moneyness: np.ndarray
    standardised_moneyness: np.ndarray



@dataclass(config=Config)
class RawOptions(BaseData):
    ivol: np.ndarray
    delta: np.ndarray
    delta_mag: np.ndarray
    gamma: np.ndarray
    vega: np.ndarray
    theta: np.ndarray
    rho: np.ndarray
    moneyness: np.ndarray
    log_moneyness: np.ndarray
    standardised_moneyness: np.ndarray
    OTM: np.ndarray
    call_flag: np.ndarray
    valid_price: np.ndarray
    underlying_px: np.ndarray
    
    def getDataKwargs(self):
        return {field: getattr(self, field) for field in self.__dataclass_fields__}

    def update_IVOL(self, idx, implied_volatility):
        self.ivol[idx] = implied_volatility
    
    def update_all_metrics_name(self, ticker, price, ivol, delta, delta_mag, gamma, vega, theta, rho, moneyness,
                                log_moneyness, standardised_moneyness, OTM, call_flag, underlying_px, valid_price):
        idx = self.name_index_map[ticker]
        self.price[idx] = price
        self.ivol[idx] = ivol
        self.delta[idx] = delta
        self.delta_mag[idx] = delta_mag
        self.gamma[idx] = gamma
        self.vega[idx] = vega
        self.theta[idx] = theta
        self.rho[idx] = rho
        self.moneyness[idx] = moneyness
        self.log_moneyness[idx] = log_moneyness
        self.standardised_moneyness[idx] = standardised_moneyness
        self.OTM[idx] = OTM
        self.call_flag[idx] = call_flag
        self.valid_price[idx] = valid_price
        self.underlying_px[idx] = underlying_px
    
    def update_all_metrics(self, idx, jdx, ivol, delta, delta_mag, gamma, vega, moneyness,
                           log_moneyness, standardised_moneyness, OTM, valid_price, underlying_px):
        self.ivol[idx] = ivol[jdx]
        self.delta[idx] = delta[jdx]
        self.delta_mag[idx] = delta_mag[jdx]
        self.gamma[idx] = gamma[jdx]
        self.vega[idx] = vega[jdx]
        self.moneyness[idx] = moneyness
        self.log_moneyness[idx] = log_moneyness
        self.standardised_moneyness[idx] = standardised_moneyness[jdx]
        self.OTM[idx] = OTM
        self.valid_price[idx]=valid_price
        self.underlying_px[idx]=underlying_px
    
    def update_with_instrument_object(self, idx, option_object, ivol, update_list):
            self.ivol[idx] = ivol
            for metric in update_list:    
                getattr(self, metric)[idx] = getattr(option_object, metric)
                
@dataclass(config=Config3)
class BaseDataClass(ABC):
    px_type: str
    x: np.ndarray 
    y: np.ndarray 
    z: np.ndarray   
    metric_axis_engine: "axis_utils.MetricAxisEngine"
    
    x_min: float = Field(init=False)
    x_max: float = Field(init=False)
    y_min: float = Field(init=False)
    y_max: float = Field(init=False)
    z_min: float = np.nan
    z_max: float = np.nan
    valid_values: bool = False
    
    def __post_init__(self):
        self.calculate_domain(self.x, self.y)
        self.z_min, self.z_max = np.nan, np.nan 
    
    def calculate_domain(self, x, y):
        if x.size > 0:
            self.x_min, self.x_max = np.min(x), np.max(x)
        else:
            self.x_min, self.x_max = np.nan, np.nan 
        if y.size > 0:
            self.y_min, self.y_max = np.min(y), np.max(y)
        else:
            self.y_min, self.y_max = np.nan, np.nan 

    def get_limits(self):
        return self.x_min, self.x_max, self.y_min, self.y_max, self.z_min, self.z_max

    def getData(self):
        return self.x, self.y, self.z
    
    def get_plotitem_args(self):
        return [self.px_type, self] 

    def update_from_raw(self, raw):
        x, y, z, _ = self.metric_axis_engine.transform_values(raw)
        x, y, z = utils.filter_nans_on_z(x, y, z)
        self.update_data(x, y, z)
    
    def setData(self, x, y, z):
        self.x, self.y, self.z = x, y, z    
    
    @abstractmethod
    def update_data(self, x, y, z):
        pass


@dataclass(config=Config3)
class Scatter(BaseDataClass):
    
    def calculate_limits(self, x, y, z):
        self.calculate_domain(x, y)
        if any(~np.isnan(z)):
            self.z_min, self.z_max = np.nanmin(z), np.nanmax(z)
        else:
            self.z_min, self.z_max = np.nan, np.nan
    
    def update_data(self, x, y, z):
        if z.size == 0:
            self.valid_values=False
        else:        
            self.x, self.y, self.z = x, y, z
            self.valid_values=True
            self.calculate_limits(self.x, self.y, self.z)
            
    def getDataKwargs(self):
        x, y, z = super().getData()
        return {"pos" :np.column_stack((x, y, z))}

@dataclass(config=Config3)
class Surface(BaseDataClass):
    interpolator: Any = None
    n_x: int = 50
    n_y: int = 50
    last_interpolation_attempt_valid: bool = Field(init=False)
    
    def __post_init__(self):   
        self.create_interpolation_domain(self.x.min(),
                                         self.x.max(),
                                         self.y.min(),
                                         self.y.max(),
                                         )
        self.z = np.full((self.n_x, self.n_y), np.nan)
        self.calculate_domain(self.x, self.y)

    def getDataKwargs(self):
        return {"x" : self.x, "y" : self.y, "z" : self.z}

    def calculate_limits(self, x, y, z):
        self.calculate_domain(x, y)
        if np.any(~np.isnan(z)):
            self.z_min, self.z_max = np.nanmin(z), np.nanmax(z)
        else:
            self.z_min, self.z_max = np.nan, np.nan

    def verify_limits(self, x_min, x_max, y_min, y_max):
        if (self.x_min != x_min
            or self.x_max != x_max
            or self.y_min != y_min
            or self.y_max != y_max):

            self.x_min = x_min 
            self.x_max = x_max
            self.y_min = y_min
            self.y_max = y_max
            self.create_interpolation_domain(x_min, x_max, y_min, y_max)
    
    def invalid_surface_cleanup(self):
        self.valid_values=False
        self.z *= np.nan
        
    def set_interpolation_domain(self, x, y):
        self.x = x
        self.y = y
        
    def create_interpolation_domain(self, x_min, x_max, y_min, y_max):
        self.x = np.linspace(x_min, x_max, self.n_x)
        self.y = np.linspace(y_min, y_max, self.n_y)
             
    def _interpolate_surface(self, x, y, z):
        try:
            self.interpolator.fit(x, y, z)                
        except Exception as e:
            print(f"The interpolator could not fit: {e}")
            self.invalid_surface_cleanup()
            self.last_interpolation_attempt_valid=False
            return
        
        try:
            self.z = self.interpolator.evaluate(self.x, self.y)

        except Exception as e:
            print(f"The interpolator could not evaluate: {e}")
            self.invalid_surface_cleanup()
            self.last_interpolation_attempt_valid=False
            return
        
        self.last_interpolation_attempt_valid=True
        self.valid_values = True
        self.z_min, self.z_max = np.min(z), np.max(z)
                
        if not self.last_interpolation_attempt_valid:
            print("The last interpolation was valid")

    def update_data(self, x, y, z):
        x, y, z = utils.filter_nans_on_z(x, y, z)
        if z.size < 5:
            warnings.warn(exceptions.InsufficientDataWarning(f"Unable to interpolate: z has only {z.size} points (min 5 required)", 
                                                            code=100),
                        stacklevel=2  # Point to the caller's line
                        )
            self.last_interpolation_attempt_valid = False
            self.invalid_surface_cleanup()
            return
        else:
            self.verify_limits(np.min(x), np.max(x), np.min(y), np.max(y))
            self._interpolate_surface(x, y, z)


@dataclass(config=Config3)
class Slices(   ):
    surface: Surface = None
    surface_axis_directions: str = ""
    base_domain: InitVar[BaseDomain] = None
    scatter: Optional[Scatter] = Field(init=True)
    displayed_slices: List[float] = Field(init=False)
    slice_container: Dict[float, List[np.ndarray]] = Field(init=False)
    n_x: int = Field(init=False)
    domain_vec: np.ndarray = Field(init=False)
    domain_idx_map: Dict[str, int] = Field(init=False)
        
    def __post_init__(self, base_domain):
        if self.surface_axis_directions == "xz":
            self.domain_vec = np.unique(base_domain.expiry)
            domain_matrix = self.x
            self.n_x = self.surface.y.size
        else:
            self.domain_vec = np.unique(base_domain.strike)
            domain_matrix = self.y
            self.n_x = self.surface.x.size
            
        if self.domain_vec.size > 15:
            self.filter_domain(base_domain)

        self.domain_idx_map = {val: idx for idx, val in enumerate(self.domain_vec)}
        self.slice_container={}

        for x_val in self.domain_vec:
            self.slice_container[x_val] = np.full(self.n_x, np.nan)
        
    def filter_domain(self, base_domain):
        value_dict = {}
        for value in self.domain_vec:
            if self.surface_axis_directions == "xz":
                mask = base_domain.expiry == value
                filtered_domain = base_domain.expiry[mask]
            else:
                mask = base_domain.strike == value
                filtered_domain = base_domain.strike[mask]
            value_dict[value] = len(filtered_domain)
        top_12 = sorted(value_dict, key=value_dict.get, reverse=True)[:12]
        top_12 = np.sort(top_12)
        self.domain_vec = self.domain_vec[np.isin(self.domain_vec, top_12)]

    
    def update_data(self, x, y, z):
        if not self.scatter is None:
            self.scatter.update_data(x, y, z)
        self.surface.update_data(x, y, z)

        if not self.valid_values:
            return 
        else:
            self.update_slice(x, y, z)
    
    def _interpolate_slice(self, dependent_values, constant_values):
        if self.surface_axis_directions == "xz":
            z_interp = self.interpolator.evaluate(dependent_values, constant_values)
        else:
            z_interp = self.interpolator.evaluate(constant_values, dependent_values)

        if z_interp.ndim != 1:
            if z_interp.shape[1] == 1 or z_interp.shape[0] == 1:
                z_interp = z_interp.flatten()
            else:
                if self.surface_axis_directions=="yz":
                    z_interp = z_interp[0,:]
                else:
                    z_interp = z_interp[:,0]
        return z_interp

    def update_slice(self, x, y, z):
        dependent_values = x if self.surface_axis_directions == "xz" else y
        size = dependent_values.size
        constant_values = [np.nan] * size
        for constant_value in self.displayed_slices:
            constant_values[:] = [constant_value] * size
            z_interp = self._interpolate_slice(dependent_values, constant_values)        





def get_settings_stylesheets(style_sheet):
    match style_sheet:

        case "QPushButton":
            style_sheet_str = """
                        QPushButton {background-color: #464646;
                                    color: white;
                                    border: 1px solid black;
                                    padding: 5px;
                                    }
                        QPushButton:checked {background-color: #9e9e9f;
                                            color: black;
                                            }
                    """
        case "QComboBox":
            style_sheet_str = """

                            QComboBox {
                                background-color: #464646;
                                color: white;
                            }
                            QComboBox::drop-down {
                                border-color: #464646;
                                background-color: #464646;
                            }
                            QComboBox::item {
                                background-color: #464646;
                                color: white; 
                            }
                            QComboBox::item:selected {
                                background-color:  #9e9e9f;
                                color: black; 
                            }
                            QComboBox QAbstractItemView {
                                background-color: #464646;
                                color: white; 
                            }
                            """
                
        case "QMenu":
            style_sheet_str = """
                            QMenu::item {
                                        background-color: #464646;
                                        color: white
                                        }
                            QMenu::item:selected {
                                                background-color: #9e9e9f; 
                                                color: black;
                                                }
                            """


        case "QToolButton":
            style_sheet_str = """
                                QToolButton {
                                            color: white;
                                            background-color: #464646;
                                            }
                                            
                                QToolButton::selected {
                                                        color: black;
                                                        background-color: #9e9e9f;
                                                        }

                                QToolButton::hover {
                                                        color: black;
                                                        background-color: #9e9e9f;
                                                        }

                                """
        case "QTableWidget":
            style_sheet_str = """
                            QTableView {
                                        border-style: none;
                                        background-color: black;
                                        padding: 0px;
                                        margin: 0px;
                                        color: white;
                                        border-left: 2px solid #414141;
                                        border-right: 2px solid #414141;
                                        selection-background-color: #9e9e9f;
                                        gridline-color: black;
                                        }
                            QTableWidget::item {
                                                padding: 0px;
                                                margin: 0px;
                                                border-bottom: 1px solid black;  
                                                border-top: 1px solid black;
                                                }
                            QTableWidget::item:selected {
                                                        background-color: #9e9e9f;
                                                        padding: 0px;
                                                        margin: 0px;
                                                        color: white;
                                                        border-bottom: 1px solid black;  
                                                        border-top: 1px solid black;
                                                        }
                            """                   
            
        case "SubQTableWidget":
            style_sheet_str = """
                QTableView {
                    border-style: none;
                    background-color: black;
                    padding: 0px;
                    margin: 0px;
                    color: white;
                    border-left: 2px solid #414141;
                    border-right: 2px solid #414141;
                    selection-background-color: #9e9e9f;
                }
                QTableWidget::item {
                    padding: 0px;
                    margin: 0px;
                    border-top: 1px solid #414141;
                    border-bottom: 1px solid #414141;

                }
                QTableWidget::item:selected {
                    background-color: #414141;
                    padding: 0px;
                    margin: 0px;
                    color: white;
                }
            """                   
            
        case "QLabel":
            style_sheet_str = """background-color : #232323;
                                color : white;
                                padding: 0px;
                                margin: 0px; 
                                
                                """
        
        
        case "QLineEdit":
            style_sheet_str = """
                            QLineEdit {
                                    background-color: #fb8b1e;
                                    color: black;
                                }
                            QLineEdit:focus {
                                    background-color: #fb8b1e;
                                    border: 2px solid black;
                                            }
                            QLineEdit::selection {
                                                    background-color:  lightblue;
                                                    color: #ffffff;                                  
                                                    }
                            """
        case "SpotQLabel":
            style_sheet_str = """
                            QLabel {
                                        background-color: black;
                                        color: #fb8b1e;
                                    }
                              """
                              
        case "StrikeOptionsComboBox":
            style_sheet_str = """
                            QComboBox {
                                background-color: #fb8b1e;
                                color: black;
                                border: None;
                                text-align: right;
                            }
                            QComboBox:focus {
                                background-color: #fb8b1e;
                                border: 2px solid black;
                                text-align: right;
                            }
                            QComboBox::drop-down {
                                border-color: black;
                                background-color: #fb8b1e;
                                text-align: right;
                            }
                            QComboBox::item {
                                background-color: #fb8b1e;
                                color: black;
                                text-align: right;
                            }
                            QComboBox::item:selected {
                                background-color: #d97d1a; 
                                color: black;
                                text-align: right;
                            }
                            QComboBox QAbstractItemView {
                                background-color: #fb8b1e;
                                color: black;
                                border: None;
                                text-align: right;
                            }
                        """
        
    return style_sheet_str


import sys
import time
import threading
from PySide6 import QtCore, QtWidgets
import pyqtgraph as pg
from PySide6.QtCore import QThread, QObject, Slot


class ClickablePlot(pg.PlotWidget):
    # Define a custom signal that emits the mouse click position (QPointF)
    mouseClicked = QtCore.Signal(object)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMouseEnabled(x=False, y=False)
        self.scene().sigMouseClicked.connect(self._on_mouse_click)

    def _on_mouse_click(self, event):
        if event.button() == QtCore.Qt.MouseButton.LeftButton:
            pos = event.scenePos()
            if self.sceneBoundingRect().contains(pos):
                mousePoint = self.getPlotItem().vb.mapSceneToView(pos)
                self.mouseClicked.emit(mousePoint)


class Worker1(QObject):
    @Slot(object)
    def run(self, pos):
        time.sleep(.5)
        print(f"Handler 1: {pos} | Thread: {threading.current_thread().name}")


class Worker2(QObject):
    @Slot(object)
    def run(self, pos):
        time.sleep(.5)
        print(f"Handler 2: {pos} | Thread: {threading.current_thread().name}")


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("Mouse Click Signal Example")
        self.resize(800, 600)

        # Create and set up plot widget
        self.plot = ClickablePlot()
        self.setCentralWidget(self.plot)

        print(f"MainWindow thread: {threading.current_thread().name}")

        self.thread1 = QThread()
        self.worker1 = Worker1()
        self.worker1.moveToThread(self.thread1)
        self.plot.mouseClicked.connect(self.worker1.run)
        self.thread1.start()

        self.thread2 = QThread()
        self.worker2 = Worker2()
        self.worker2.moveToThread(self.thread2)
        self.plot.mouseClicked.connect(self.worker2.run)
        self.thread2.start()


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    win = MainWindow()
    win.show()
    sys.exit(app.exec())



from PySide6 import QtCore, QtGui, QtWidgets

def get_global(key):
    match key:
        case "font":
            ret = QtGui.QFont("Neue Haas Grotesk Text Pro", 10)
        
        case "widgets":
            ret = """
                  QPushButton {background-color: #fb8b1e;
                                color: black;
                                border: none;           
                                padding: 5px 10px;   
                  }
                  QPushButton:hover {background-color: #e67e17; 
                  }
                  QPushButton:pressed {background-color: #d35400;
                  }
                  """
    return ret


import sys
import numpy as np
from PySide6 import QtCore, QtGui, QtWidgets
import pyqtgraph.opengl as gl
from OpenGL import GL, GLU




class GroupedGLTextItem(gl.GLGraphicsItem.GLGraphicsItem):
    """
    A custom GLGraphicsItem for efficiently drawing multiple text labels.

    This item manages and renders a collection of text strings at specified
    3D positions, optimizing the rendering process by handling them as a
    single unit.
    """

    def __init__(self, **kwds):
        super().__init__()
        self.texts = []
        self.positions = np.empty((0, 3))
        self.font = QtGui.QFont('Helvetica', 16)
        self.color = QtGui.QColor(QtCore.Qt.white)

    def setData(self, positions, texts, color=None, font=None):
        """
        Sets the data for the text items.

        Args:
            positions (np.ndarray): A NumPy array of shape (N, 3) specifying the
                                    x, y, z coordinates for each text string.
            texts (list): A list of N strings to be displayed.
            color (QtGui.QColor, optional): The color of the text. Defaults to white.
            font (QtGui.QFont, optional): The font to use for the text.
        """
        self.positions = np.asarray(positions)
        self.texts = texts
        if color is not None:
            self.color = color
        if font is not None:
            self.font = font
        self.update()

    def paint(self):
        """
        The paint method called by the GLViewWidget to render the text.
        """
        if len(self.texts) == 0:
            return

        self.setupGLState()

        painter = QtGui.QPainter(self.view())
        painter.setPen(self.color)
        painter.setFont(self.font)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        viewport = GL.glGetIntegerv(GL.GL_VIEWPORT)

        modelview_matrix = GL.glGetDoublev(GL.GL_MODELVIEW_MATRIX)
        projection_matrix = GL.glGetDoublev(GL.GL_PROJECTION_MATRIX)
        for i in range(len(self.texts)):
            pos = self.positions[i]
            text = self.texts[i]
            """
            gl_pos = modelview * QtGui.QVector4D(pos[0], pos[1], pos[2], 1.0)
            gl_pos = projection * gl_pos
            print(f"\ngl_pos\n")
            print(gl_pos)
            
            w = self.view().width()
            h = self.view().height()
            
            x = (gl_pos.x() / gl_pos.w() + 1.0) / 2.0 * w
            y = (1.0 - (gl_pos.y() / gl_pos.w())) / 2.0 * h
            """
            
            
            screen_pos = GLU.gluProject(pos[0], pos[1], pos[2],
                                       modelview_matrix,
                                       projection_matrix,
                                       viewport)

            x = screen_pos[0]
            y = self.view().height() - screen_pos[1]
            painter.drawText(QtCore.QPointF(x, y), text)
        painter.end()
 

class MainWindow(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Optimized 3D Text Rendering")
        self.setGeometry(100, 100, 800, 600)

        central_widget = QtWidgets.QWidget()
        self.setCentralWidget(central_widget)
        layout = QtWidgets.QVBoxLayout(central_widget)

        self.view = gl.GLViewWidget()
        layout.addWidget(self.view)

        # Add a grid to the scene for context
        grid = gl.GLGridItem()
        self.view.addItem(grid)

        # Create and add the grouped text item
        self.grouped_text_item = GroupedGLTextItem()
        self.view.addItem(self.grouped_text_item)

        # Initial data for the text items
        self.positions = np.array([
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1]
        ])
        self.texts = ["Text 1", "Text 2", "Text 3"]
        
        self.grouped_text_item.setData(positions=self.positions, texts=self.texts)

        # Timer to update the text and positions
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update_text)
        #self.timer.start(50)  # Update every 50ms

        self.frame_count = 0

    def update_text(self):
        """
        Updates the positions and text of the labels.
        """
        # Animate the positions
        angle = np.radians(self.frame_count * 2)
        rotation_matrix_z = np.array([
            [np.cos(angle), -np.sin(angle), 0],
            [np.sin(angle), np.cos(angle), 0],
            [0, 0, 1]
        ])
        
        new_positions = np.dot(self.positions, rotation_matrix_z.T)

        # Update the text
        new_texts = [f"Frame: {self.frame_count}" for _ in self.texts]

        # Update the data in the grouped item
        self.grouped_text_item.setData(positions=new_positions, texts=new_texts)
        
        self.frame_count += 1

if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    window = MainWindow()
    window.show()
    sys.exit(app.exec())


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from instruments.instruments import Option, Spot
    from ..graphics_items.GL3DGraphicsItems import GL3DLinePlotDataItem
    from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
    from uuid import uuid4 

from PySide6 import QtCore, QtGui
import numpy as np
import numpy as np
from dataclasses import dataclass, InitVar, field
from PySide6 import QtCore
from ..graphics_items import GL3DAxisItem
from .. import utils
from pprint import pprint   




@dataclass
class ItemLimitContainer:
    axes: str
    item_ids: List[str] = field(init=False, default_factory=list)
    min_sorted: Dict[str, List[float]] = field(init=False, default_factory=dict)
    max_sorted: Dict[str, List[float]] = field(init=False, default_factory=dict)

    min_items: Dict[str, Dict[uuid4, float]] = field(init=False, default_factory=dict)
    max_items: Dict[str, Dict[uuid4, float]] = field(init=False, default_factory=dict)
    
    def __post_init__(self):
        self.min_sorted = {ax : [] for ax in self.axes}
        self.max_sorted = {ax : [] for ax in self.axes}
        self.min_items = {ax : {} for ax in self.axes}
        self.max_items = {ax : {} for ax in self.axes}
        
    def _calculate_limits(self, values):
        if np.any(~np.isnan(values)):
            min, max = np.nanmin(values), np.nanmax(values)
        else:
            min, max = np.inf, -np.inf 
        return [min, max]
        
    def limitsFromAxis(self, axis):
        return self.min_sorted[axis][0], self.max_sorted[axis][-1]
    
    def limits(self):
        return [list(self.limitsFromAxis(idx)) for idx in range(3)]

    def addItem(self, item: GL3DLinePlotDataItem):
        self.item_ids.append(item.id())
        for ax in range(3):
            min_val, max_val = item.dataset.limits_subset[ax]
            self.min_items[ax][item.id()]=min_val
            self.max_items[ax][item.id()]=max_val
            self._sort_limits(ax)

    def _sort_limits(self, axis):
        self.min_sorted[axis]=list(self.min_items[axis].values())
        self.min_sorted[axis].sort()
        self.max_sorted[axis]=list(self.max_items[axis].values())
        self.max_sorted[axis].sort()
    
    def updateAxisFromItem(self, item: GL3DLinePlotDataItem, axis):
        min_val, max_val = item.dataset.limits_subset[axis]
        self.min_items[axis][item.id()]=min_val
        self.max_items[axis][item.id()]=max_val
        self._sort_limits(axis)

    def removeItem(self, internal_id):
        if internal_id in self.item_ids:
            self.item_ids.remove(internal_id)
            del self.min_sorted[internal_id]
            del self.max_sorted[internal_id]
            del self.min_items[internal_id]
            del self.max_items[internal_id]
            
        for ax in range(3):
            self._sort_limits(ax)


@dataclass(slots=True)
class AxisNormaliser:
    axis: str
    axis_str: int
    padding: float
    min: float = field(init=False,default=np.nan)
    max: float = field(init=False,default=np.nan)
    min_UB: float = field(init=False,default=np.nan)
    max_LB: float = field(init=False,default=np.nan)
    
    def limits(self):
        return self.min, self.max
    
    def normalise(self, values):
        return (values - self.min) / (self.max - self.min)
    
    def unnormalise(self, values):
        return values * (self.max - self.min) + self.min
    
    def updateFromItemLimit(self, limit: List[float, float]):
        self.min = limit[0] - self.padding * (limit[1] - limit[0])
        self.max = limit[1] + self.padding * (limit[1] - limit[0])
        self._computeMinUB()
        self._computeMaxLB()
    
    def setMax(self, max):
        self.max = max
        self._computeMaxLB()
    
    def _computeMinUB(self):
        if not np.isnan(self.max):
            d = self.max - self.min
            self.min_UB = self.min + 0.4 * d
            
    def _computeMaxLB(self):
        if not np.isnan(self.min):
            d = self.max - self.min
            self.max_LB = self.max - 0.4 * d

    def checkLimit(self, limit: List[float, float]):
        flag=False
        if limit[0] < self.min or limit[1] > self.min_UB:
            flag=True
        if limit[1] > self.max or limit[0] < self.max_LB:
            flag=True
        return flag
    
    def mask_from_data(self, data):
        return (data >= self.min) & (data <= self.max)
    
        
class ViewBox(QtCore.QObject):
    sigXRangeChanged = QtCore.Signal(QtCore.QObject)
    sigYRangeChanged = QtCore.Signal(QtCore.QObject)
    sigZRangeChanged = QtCore.Signal(QtCore.QObject)
    
    sigViewChanged = QtCore.Signal(QtCore.QObject)

    #sigXNormChanged = QtCore.Signal(AxisNormaliser)
    #sigYNormChanged = QtCore.Signal(AxisNormaliser)
    #sigZNormChanged = QtCore.Signal(AxisNormaliser)

    def __init__(self,
                 norm_dims=None,
                 padding=[0, 0, 0]
                 ):
        super().__init__()
        self.item_limit_container = ItemLimitContainer([idx for idx in range(3)])
        
        if norm_dims is None:
            self.norm_dims = {ax : 1 for ax in range(3)}
        else:
            self.norm_dims = norm_dims
        self.padding=padding
        self._recursive_flag_container={}
        self.axis_normaliser_container = {ax_idx : AxisNormaliser(ax_idx, ax_str, pad_val) for ax_idx, (ax_str, pad_val) in enumerate(zip("xyz", self.padding))}
        self._viewRange = [[0, 0] for _ in range(3)]
        self.addedItems=[]
        self.state={
                    "autoRange" : [True]*3,
                    "aspectLocked" : False,
                    }
        self._idx_str_map = {i : s for i, s in enumerate("xyz")}        
    
    def addItem(self,
                item: GLGraphicsItem,
                ignoreBounds: bool=False
                ):
        if hasattr(item, "type"):
            if item.type=="axis":
                return
        if isinstance(item, GL3DAxisItem.GL3DViewBox):
            getattr(self, f"sig{self._idx_str_map[item.direction].upper()}RangeChanged").connect(item.update_values)
        else:
            if not ignoreBounds and hasattr(item, "setData") and callable(item.setData):
                self._recursive_flag_container[item.id()]=False 
                
                item.linktoView(self)
                self.blockSignals(True)
                self.connect_slot_to_norm_sig(item.id(), item.setDataFromView)
                self.connect_item_to_rangechanged_sig(item)
                self.blockSignals(False)
                self.addedItems.append(item)
                if not ignoreBounds:
                    self._addItemLimitCheck(item)
                item
        return item

    def removeItem(self, item):
        if item in self.addedItems:
            self.addedItems.remove(item)
            self.item_limit_container.removeItem(item)
    
    def normaliseToView(self, x, y, z):
        return [self.axis_normaliser_container[0].normalise(x),
                self.axis_normaliser_container[1].normalise(y),
                self.axis_normaliser_container[2].normalise(z)]
    
    def normaliseToView2(self, x=None, y=None, z=None):
        values_normalised = []
        for idx, arr in enumerate([x, y, z]):
            if not arr is None:
                val_norm = self.axis_normaliser_container[idx].normalise(arr)
                values_normalised.append(val_norm)
        return values_normalised

    def resetNormalisation(self):
        self.axis_normaliser_container = {axis : AxisNormaliser(axis) for axis in range(3)}
        self.item_limit_container = ItemLimitContainer([idx for idx in range(3)])
        
    def updateNormalisation2222222(self, item, pos=None, x=None, y=None, z=None):
        if not pos is None:
            x=pos[:,0]
            y=pos[:,1]
            z=pos[:,2]
        
        for axis, arr in zip(range(3), (x,y,z)):
            self.item_limit_container.updateAxisFromItem(item, axis)
            self._check_view_box_limits(axis, item.id(), ignore_update_signals=True)
    
    def viewRange(self) -> List[List[np.ndarray]]:
        return [vr[:] for vr in self._viewRange] 
    
    def _ProcessLimitUpdate(self, item_id, ax, limits):
        plot_limits = self.axis_normaliser_container[ax]
        plot_limits.updateFromItemLimit(limits)
        self._viewRange[ax][:] = plot_limits.limits()
        self._recursive_flag_container[item_id]=True
        self.sigViewChanged.emit(self)
        self._recursive_flag_container[item_id]=False

    def _check_view_box_limits(self, ax, item_id, ignore_update_signals=False):
        if self.state[]=="OnBoundary":
            item_limits=self.item_limit_container.limitsFromAxis(ax)
            plot_view_item=self.axis_normaliser_container[ax]
            if all((not np.isinf(item_limits).any(),
                    not np.isnan(item_limits).any()
                    )):

                new_limit = plot_view_item.checkLimit(item_limits)
                if new_limit: 
                    self._ProcessLimitUpdate(item_id, ax, item_limits)

    def _addItemLimitCheck(self, item):
        self.item_limit_container.addItem(item)
        for axis in range(3):
            limits = self.item_limit_container.limitsFromAxis(axis)
            self._ProcessLimitUpdate(item.id(), axis, limits)
        
    def _preNormalisation(self, item: GL3DLinePlotDataItem, axis, values):
        self.item_limit_container.updateAxisFromItem(item, axis)
        self._check_view_box_limits(axis, item.id())
            
    def normalise(self, axis, item, values, ignoreBounds=False):
        if not ignoreBounds:
            if not self._recursive_flag_container[item.id()]:
                self._preNormalisation(item, axis, values)
        return self.axis_normaliser_container[axis].normalise(values)

    def connect_slot_to_norm_sig(self, internal_name, slot):
        self._recursive_flag_container[internal_name]=False
        self.blockSignals(True)
        self.sigViewChanged.connect(slot)
        #for ax in "XYZ":
            #getattr(self, f"sig{ax}NormChanged").connect(slot)
        self.blockSignals(False)
        
    def connect_item_to_rangechanged_sig(self, item):
        self._recursive_flag_container[item.id()]=False
        self.blockSignals(True)
        for ax in "XYZ":
            getattr(self, f"sig{ax}RangeChanged").connect(getattr(item, f"update{ax}View"))
        self.blockSignals(False)
    
    def setRange(self, xRange=None, yRange=None, zRange=None):
        if all((xRange is None,
                yRange is None,
                zRange is None)):
            for axis in range(3):
                min_val, max_val = self.item_limit_container.limitsFromAxis(axis)
                getattr(self, f"sig{axis.upper()}RangeChanged").emit(AxisNormaliser(axis, [0,1,2].index(axis), min_val, max_val), self)
            return
        x=False
        if not xRange is None:
            self.axis_normaliser_container[0].updateFromItemLimit(xRange)
            self.updateViewRange(0)
            x=True
            
        y=False
        if not yRange is None:
            self.axis_normaliser_container[1].updateFromItemLimit(yRange)
            self.updateViewRange(1)
            y=True
        
        z=False
        if not zRange is None:
            self.axis_normaliser_container[2].updateFromItemLimit(zRange)
            self.updateViewRange(2)
            z=True
        
        for item in self.addedItems:
            for ax, flag in enumerate((x,y,z)):
                if flag:
                    self.item_limit_container.updateAxisFromItem(item, ax)
        
        self.sigViewChanged.emit(self)
        
    def updateViewRange(self, ax):
        self._viewRange[ax][:]=self.axis_normaliser_container[ax].limits()
        getattr(self, f"sig{self._idx_str_map[ax].upper()}RangeChanged").emit(self)
    
                
    def resetToDataLimits(self):
        for axis in range(3):
            min_val, max_val = self.item_limit_container.limitsFromAxis(axis)
            self.setRange(**{f"{axis}Range": [min_val, max_val]})


from collections import deque, defaultdict
from typing import Dict, List, Set, Tuple, Optional, NamedTuple

class RequirementSpec(NamedTuple):
    """
    Specification for a calculation requirement
    - metric: The financial metric to calculate
    - px_type_option: px type for the option (bid/ask/mid)
    - px_type_underlying: px type for the underlying (bid/ask/mid)
    """
    metric: str
    px_type_option: Optional[str] = None
    px_type_underlying: Optional[str] = None

    def __repr__(self):
        parts = []
        if self.px_type_option:
            parts.append(f"option:{self.px_type_option}")
        if self.px_type_underlying:
            parts.append(f"underlying:{self.px_type_underlying}")
        return f"{self.metric}" + (f" ({', '.join(parts)})" if parts else "")

class RequirementsTracker:
    # Dependency relationships between metrics
    DEPENDENCIES: Dict[str, List[str]] = {
                                        "px": [],
                                        "underlying_px": [],
                                        "ivol": ["px", "underlying_px"],
                                        "delta": ["ivol"],
                                        "gamma": ["ivol"],
                                        "vega": ["ivol"],
                                        "theta": ["ivol"],
                                        "rho": ["ivol"],
                                        "standardised_moneyness": ["ivol"],
                                        "moneyness": ["underlying_px"],
                                        "log_moneyness": ["underlying_px"],
                                        "forward_moneyness": ["underlying_px"],
                                        "OTM": ["underlying_px"],
                                        "valid_px": ["px"]
                                    }
    
    # Metrics that use option px type
    OPTION_px_METRICS = {"px",
                        "ivol",
                        "delta",
                        "gamma",
                        "vega",
                        "theta",
                        "rho", 
                        "standardised_moneyness",
                        "valid_px"
                        }
    
    # Metrics that use underlying px type
    UNDERLYING_px_METRICS = {"underlying_px", "ivol", "delta", "gamma", "vega", "theta", 
                               "rho", "standardised_moneyness", "moneyness", "log_moneyness", 
                               "forward_moneyness", "OTM"}

    def __init__(self):
        """Initialize with empty requirements"""
        # Tracks all requirements including dependencies
        self.current_requirements: Set[RequirementSpec] = set()
        
        # Tracks explicitly added requirements
        self.explicit_requirements: Set[RequirementSpec] = set()
        
        # Tracks requirements by metric
        self.requirements_by_metric: Dict[str, Set[RequirementSpec]] = defaultdict(set)

    def add_interface_requirement(self, requirement: RequirementSpec):
        """
        Add a requirement for a specific interface
        Resolves and stores all necessary dependencies
        """
        self.explicit_requirements.add(requirement)
        
        # BFS to resolve dependencies
        queue = deque([requirement])
        visited = set()
        
        while queue:
            current = queue.popleft()
            
            # Skip already processed requirements
            if current in visited:
                continue
            visited.add(current)
            
            # Add to current requirements if new
            if current not in self.current_requirements:
                self.current_requirements.add(current)
                self.requirements_by_metric[current.metric].add(current)
                
                # Process dependencies
                for dep_metric in self.DEPENDENCIES.get(current.metric, []):
                    # Create dependency requirement with inherited px types
                    dep_req = self._create_dependency_requirement(
                        dep_metric, 
                        current.px_type_option,
                        current.px_type_underlying
                    )
                    queue.append(dep_req)

    def _create_dependency_requirement(self, metric: str, 
                                       option_pt: Optional[str], 
                                       underlying_pt: Optional[str]
                                       ) -> RequirementSpec:
        """Create requirement for a dependency with inherited px types"""
        # Inherit option px type if metric uses it
        inherit_option_pt = option_pt if metric in self.OPTION_px_METRICS else None
        
        # Inherit underlying px type if metric uses it
        inherit_underlying_pt = underlying_pt if metric in self.UNDERLYING_px_METRICS else None
        
        return RequirementSpec(metric=metric,
                               px_type_option=inherit_option_pt,
                               px_type_underlying=inherit_underlying_pt
                               )

    def remove_interface_requirement(self, requirement: RequirementSpec):
        """
        Remove a requirement and its dependencies
        (if not used by other requirements)
        """
        if requirement not in self.explicit_requirements:
            return
            
        self.explicit_requirements.remove(requirement)
        
        # Find all requirements to remove
        to_remove = set()
        queue = deque([requirement])
        
        while queue:
            current = queue.popleft()
            to_remove.add(current)
            
            # Find dependencies that might be removable
            for dep_metric in self.DEPENDENCIES.get(current.metric, []):
                # Get all possible dependency requirements
                for dep_req in self.requirements_by_metric.get(dep_metric, set()):
                    # Check if this dependency is only used by the current requirement
                    if all(
                        req in to_remove 
                        for req in self.current_requirements 
                        if dep_req in self._get_dependencies(req)
                    ):
                        queue.append(dep_req)
        
        # Remove identified requirements
        for req in to_remove:
            self.current_requirements.discard(req)
            self.requirements_by_metric[req.metric].discard(req)
            # Clean up empty metric entries
            if not self.requirements_by_metric[req.metric]:
                del self.requirements_by_metric[req.metric]

    def _get_dependencies(self, requirement: RequirementSpec) -> Set[RequirementSpec]:
        """Get direct dependencies for a requirement"""
        dependencies = set()
        for dep_metric in self.DEPENDENCIES.get(requirement.metric, []):
            dep_req = self._create_dependency_requirement(
                dep_metric,
                requirement.px_type_option,
                requirement.px_type_underlying
            )
            dependencies.add(dep_req)
        return dependencies

    def get_current_requirements(self) -> Set[RequirementSpec]:
        """Get all current requirements including dependencies"""
        return self.current_requirements.copy()

    def get_explicit_requirements(self) -> Set[RequirementSpec]:
        """Get requirements explicitly added for interfaces"""
        return self.explicit_requirements.copy()

    def get_requirements_for_metric(self, metric: str) -> Set[RequirementSpec]:
        """Get all requirements for a specific metric"""
        return self.requirements_by_metric.get(metric, set()).copy()

    def get_required_px_types(self) -> Dict[str, Set[str]]:
        """Get required px types by metric type"""
        px_types = {
            "option": set(),
            "underlying": set()
        }
        
        for req in self.current_requirements:
            if req.px_type_option:
                px_types["option"].add(req.px_type_option)
            if req.px_type_underlying:
                px_types["underlying"].add(req.px_type_underlying)
                
        return px_types

    def clear_all_requirements(self):
        """Reset all requirements"""
        self.current_requirements = set()
        self.explicit_requirements = set()
        self.requirements_by_metric = defaultdict(set)

    def __repr__(self):
        return f"RequirementsTracker({len(self.current_requirements)} requirements)"


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from instruments.instruments import Option, ABCInstrument, Derivative
    from instruments.utils import InstrumentManager
    from ..data_classes.classes import Slice
    
import pyqtgraph as pg
import numpy as np
from PySide6 import QtCore, QtWidgets, QtGui
from .option_monitor import extra_widgets, stylesheets as utils_omon


class CustomTextItem(pg.TextItem):
    def __init__(self, text=None, view_box_anchor=None, **kwargs):
        self.view_box_anchor=view_box_anchor
        if self.view_box_anchor[0] == 0:
            self.use_vb_xmin=True
        else:
            self.use_vb_xmin=False
        if self.view_box_anchor[1] == 0:
            self.use_vb_ymin=False
        else:
            self.use_vb_ymin=True

        super().__init__(text=text, **kwargs)
        self.p=[self.pos()[0], self.pos()[1]]
    
    def attach_viewbox(self, vb: pg.ViewBox):
        self.vb=vb
        self.vb.sigResized.connect(self.view_box_resize)
        self.vb.sigRangeChanged.connect(self.view_box_resize)
        self.vb.sigRangeChangedManually.connect(self.view_box_resize)


    def right_click(self, *args):
        if not self.isVisible():
            super().show()
    
    def view_box_resize(self, args):
        if not self.view_box_anchor is None:
            vb_range = self.vb.viewRange()
            if self.use_vb_xmin:
                self.p[0] = vb_range[0][0]
            else:
                self.p[0] = vb_range[0][1]
            if self.use_vb_ymin:
                self.p[1] = vb_range[1][0]
            else:
                self.p[1] = vb_range[1][1]
            super().setPos(*self.p)

class CustomPlotDataItem(pg.PlotDataItem):
    def __init__(self, px_type, *args, **kwargs):
        self.px_type=px_type  
        super().__init__(*args, **kwargs)
    
    def update_from_dataclass(self, dclass: Slice):
        self.setData(**dclass.plot_item_kwargs())
    

class OptionInfoTable(QtWidgets.QWidget):
    def __init__(self, name_metric_map, getattr_map, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.getattr_map=getattr_map
        self.price_types=name_metric_map["price_types"]    
        self.cellitem_container={}
        self._on_view=True
        self._initLayout()
        self._initTables(name_metric_map, getattr_map)
        
    def _initLayout(self):
        self.setLayout(QtWidgets.QVBoxLayout())

    """
    def _initTables(self, name_metric_map, getattr_map):
        self.name_table = NameTable(name_metric_map["object"].ticker,
                                    name_metric_map["object"].underlying_px)
        
        self.metric_table = MetricTable(name_metric_map, getattr_map)
        self.layout().addWidget(self.name_table)
        self.layout().addWidget(self.metric_table)
        
        self.autoFillBackground()
        self.pop_table=False
        self.metric_table.verticalHeader().sectionResized.connect(self.update_vertical_dims)
        self.name_table.horizontalHeader().sectionResized.connect(self.update_horizontal_dims)
        self.showEvent = self.on_show_event
    """
    def update_horizontal_dims(self, *args):
        self.name_table.horizontalHeader().setFixedWidth(self.metric_table.horizontalHeader().width())
        
    def update_vertical_dims(self, *args):
        self.metric_table.verticalHeader().setFixedWidth(self.metric_table.verticalHeader().width())
        

    def on_show_event(self, event):
        self.name_table.verticalHeader().resizeSections(QtWidgets.QHeaderView.ResizeMode.ResizeToContents) 
        self.update_vertical_dims(0, 0, 0) 
        super().showEvent(event)


class MetricTable(QtWidgets.QTableWidget):
    def __init__(self, name_metric_map, getattr_map, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.getattr_map=getattr_map
        self.price_types=name_metric_map["price_types"]    
        self.nrows = 1 + len(self.getattr_map)
        self.ncols=len(self.price_types)
        self.cellitem_container={}
        self._on_view=True
        self._underlying_object = name_metric_map["object"]
        self._initTableDimensions()
        self._initCellItems(name_metric_map["object"])
        self.autoFillBackground()
        self.setShowGrid(False)
        self.pop_table=False
        
    def set_on_view(self, flag):
        self._on_view=flag
        
    def _initTableDimensions(self):
        self.verticalHeader().setVisible(True)
        self.horizontalHeader().setVisible(True)
        self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Fixed)
        self.setStyleSheet(utils_omon.get_settings_stylesheets("SubQTableWidget"))   
        
        self.setRowCount(self.nrows)
        self.setColumnCount(self.ncols)        
        
        for jdx, px_type in enumerate(self.price_types):
            col_name_item = extra_widgets.TableColumnItem(px_type)
            font = QtGui.QFont("Neue Haas Grotesk", 12)
            col_name_item.setFont(font)
            self.setHorizontalHeaderItem(jdx, col_name_item)

        underlying_label = extra_widgets.TableColumnItem("underlying_px")
        underlying_label.setTextAlignment(QtCore.Qt.AlignLeft)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        underlying_label.setFont(font)
        self.setVerticalHeaderItem(0, underlying_label)
        if self.columnCount() > 1:
            self.setSpan(0, 0, 1, self.columnCount())
        
        for idx, name in enumerate(self.getattr_map, start=1):
            row_name_item = extra_widgets.TableColumnItem(name)
            row_name_item.setTextAlignment(QtCore.Qt.AlignLeft)
            row_name_item.setFont(font)
            self.setVerticalHeaderItem(idx, row_name_item)

    def _initCellItems(self, instrument_object: ABCInstrument):
        if hasattr(instrument_object, "underlying_px"):
            value = instrument_object.underlying_px
            
            cell_item = extra_widgets.OptionMetricCellItem(f"{value:,.2f}")
            cell_item.setTextAlignment(QtCore.Qt.AlignLeft)
            self.cellitem_container["underlying_px"]=cell_item
            self.setItem(0, 0, cell_item)
            start_idx=1
        else:
            start_idx=0

        for jdx, px_type in enumerate(self.price_types):
            for idx, (name, call) in enumerate(self.getattr_map.items(), start=start_idx):
                value = getattr(instrument_object, call)(px_type)
                if name in ["delta", "gamma"]:
                    val_str=f"{value:,.2g}"
                else:
                    val_str=f"{value:,.2f}"
                cell_item = extra_widgets.OptionMetricCellItem(val_str)
                self.cellitem_container[name]=cell_item
                self.setItem(idx, jdx, cell_item)
    
    def update_view(self):
        self.update_table(self._underlying_object)
    
    def update_table(self, instrument_object: Derivative):
        if self._on_view:
            name = "underlying_px"
            cell_item = self.item(0, 0)
            cell_item.setText(f"{instrument_object.underlying_px:,.2f}")
            for jdx, px_type in enumerate(self.price_types):
                for idx, (name, call) in enumerate(self.getattr_map.items(), start=1):
                    value = getattr(instrument_object, call)(px_type)
                    if name in ["delta", "gamma"]:
                        val_str=f"{value:,.2g}"
                    else:
                        val_str=f"{value:,.2f}"
                    cell_item = self.item(idx, jdx)
                    cell_item.setText(val_str)


class OptionMetricsWithTabs(QtWidgets.QWidget):
    def __init__(self, name_metric_map, getattr_map, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.getattr_map=getattr_map
        self.instrument_names=list(name_metric_map.keys())
        self.nrows = 2 + len(self.getattr_map)
        self.metric_name_cellitem_map={}
        self.name_tab_idx_map={}
        self.buttons=[]
        self._initLayout()
        self._initTabs(name_metric_map)
        self._initTableDimensions(name_metric_map)
        self._initCellItems(name_metric_map)
        self.autoFillBackground()
        btn = self.button_group.button(0)
        btn.click()
        
    def _initLayout(self):
        self.main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.main_layout)
        self.button_group = QtWidgets.QButtonGroup()
        self.table_stack = QtWidgets.QStackedLayout()
        
        self.button_layout = QtWidgets.QHBoxLayout()
        self.button_layout.setContentsMargins(0, 0, 0, 0)
        
        self.main_layout.addLayout(self.button_layout)
        self.main_layout.addLayout(self.table_stack)
        
        self.main_layout.setStretch(0, 0)  # Buttons don't stretch vertically
        self.main_layout.setStretch(1, 1)  # Table g     
    

    def _initTabs(self, metric_object_container):
        for idx, name in enumerate(metric_object_container):
            slot=lambda idx, checked: (self.table_stack.setCurrentIndex(idx),
                                       setattr(self.table_stack.itemAt(idx).widget(), "_on_view", True)) if checked \
                                        else setattr(self.table_stack.itemAt(idx).widget(), "_on_view", False)
            
            btn=QtWidgets.QPushButton(name)
            btn.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            btn.setSizePolicy(QtWidgets.QSizePolicy.Expanding, 
                              QtWidgets.QSizePolicy.Fixed)
                        
            if len(metric_object_container) > 1:
                btn.customContextMenuRequested.connect(self._open_button_context)
            
            btn.setCheckable(True)
            btn.setChecked(False)
            self.buttons.append(btn)
            
            self.button_layout.addWidget(btn)
            #self.button_layout.addWidget(btn, stretch=1)  
            self.button_group.addButton(btn, id=idx) 
            self.button_group.idToggled.connect(slot)
            self.name_tab_idx_map[name]=idx
        self.button_layout.addStretch()
            
    def _popout_table(self, pos, button):
        ticker = button.text()
        table = self.table_stack.itemAt(self.name_tab_idx_map[ticker]).widget()
        table.set_on_view(True)
        idx = self.name_tab_idx_map[ticker]
        
        app = QtWidgets.QApplication.instance()
        table.popped=True
        
        self.button_layout.removeWidget(button)
        button.deleteLater()  
        del self.name_tab_idx_map[ticker]
        self.button_group.removeButton(button)
        
        for name, i in list(self.name_tab_idx_map.items()):
            if idx < i:
                button = self.button_group.buttons()[i-1]
                new_i = self.name_tab_idx_map[name] - 1
                self.button_group.setId(button, new_i)
                self.name_tab_idx_map[name]=new_i
 
        self.table_stack.removeWidget(table)
        table.show()
        
    def _open_button_context(self, pos):
        button = self.sender()
        menu = QtWidgets.QMenu(self)
        print_action = menu.addAction("Open in New Window")
        print_action.triggered.connect(lambda: self._popout_table([pos.x(), pos.y()], button))
        menu.exec(button.mapToGlobal(pos))
    
    def _initTableDimensions(self, metric_object_container):
        for name in self.name_tab_idx_map:
            inner_container=metric_object_container[name]
            table = MetricTable(inner_container, self.getattr_map)
            table.set_on_view(False)
            self.table_stack.addWidget(table)
            
    
    def closeEvent(self, event):
        self.parent().closingSubtable(self)
        super().closeEvent(event)
            

    def _initCellItems(self, metric_object_container):
        for name_instrument, idx in self.name_tab_idx_map.items():
            inner_container=metric_object_container[name_instrument]
            
            instrument_object=inner_container["object"]
            price_types=inner_container["price_types"]            

            value = instrument_object.underlying_px
            
            underlying_px_cell_item = extra_widgets.OptionMetricCellItem(f"{value:,.2f}")

            table = self.table_stack.itemAt(idx).widget()
            table.setItem(1, 0, underlying_px_cell_item)
            greeks_map={}
            for jdx, px_type in enumerate(price_types):
                metric_type_map = {}
                for idx, (name_metric, call) in enumerate(self.getattr_map.items(), start=2):
                    value = getattr(instrument_object, call)(px_type)
                    if name_metric in ["delta", "gamma"]:
                        val_str=f"{value:,.2g}"
                    else:
                        val_str=f"{value:,.2f}"
                    cell_item = extra_widgets.OptionMetricCellItem(val_str)
                    metric_type_map[name_metric]=cell_item
                    table.setItem(idx, jdx, cell_item)
                greeks_map[px_type] = metric_type_map
            
            self.metric_name_cellitem_map[name_instrument]={"underlying_px" : underlying_px_cell_item,
                                                            "greeks" : greeks_map}

    def updateFromOption(self, instrument_object: Option):
        if self.name_tab_idx_map[instrument_object.ticker] == self.table_stack.currentIndex():
            inner_cellitem_map = self.metric_name_cellitem_map[instrument_object.ticker]
            cell_item = inner_cellitem_map["underlying_px"]
            cell_item.setText(f"{instrument_object.underlying_px:,.2f}")
            for px_type, metric_map in inner_cellitem_map["greeks"].items():
                for name_metric, cell_item in metric_map.items():
                    value = getattr(instrument_object, self.getattr_map[name_metric])(px_type)
                    
                    if abs(value) < 0.01:
                        val_str=f"{value:,.2g}"
                    else:
                        val_str=f"{value:,.2f}"
                    cell_item.setText(val_str)
            




from . import stylesheets as stylesheets
import numpy as np
from . import extra_widgets
from datetime import datetime   
import time 
from line_profiler import profile
from . import utils as utils_omon


def create_and_set_text_widget(table, widget, string, idx, jdx):
    text_item = widget(string)
    table.setItem(idx, jdx, text_item)

def change_strikes(new_n_strikes, table, expiry_attribute):
    new_n_strikes = int(float(new_n_strikes))
    table.blockSignals(True)
    table.setUpdatesEnabled(False)
    if new_n_strikes == len(table.current_expiry_strike_map[expiry_attribute.expiry]):
        return
    else:
        table.setItemDelegate(None)
        update_strikes_for_expiry(table, new_n_strikes, expiry_attribute)
        table.setItemDelegate(extra_widgets.CustomDelegate(table))
        
    table.blockSignals(False)
    table.setUpdatesEnabled(True)
    
def create_strike_combobox(expiry_attribute):
    combobox = extra_widgets.StrikeOptionsComboBox()
    combobox.addItem(str(0))
    for n_strike, _ in enumerate(expiry_attribute.strikes, start=1):
        combobox.addItem(str(n_strike))
    return combobox

def create_expiry_section(expiry, strikes, insturment_names, metrics, start_idx):
    table_widget_items=[]
    
    table_widget_item = extra_widgets.OptionExpiryCellItem(expiry)
    table_widget_items.append({"pos" : (start_idx, 0), "item" : table_widget_item})
    
    for jdx, _ in enumerate(metrics, start=1):
        table_widget_item = extra_widgets.OptionExpiryCellItem(None)
        table_widget_items.append({"pos" : (start_idx, jdx), "item" : table_widget_item})

    for strike, ticker in zip(strikes, insturment_names):        
        start_idx += 1
        
        table_widget_item = extra_widgets.OptionNameCellItem(ticker)
        table_widget_items.append({"pos" : (start_idx, 0), "item" : table_widget_item})
        
        for jdx, _ in enumerate(metrics, start=1):     
            table_widget_item = extra_widgets.OptionMetricCellItem("")           
            table_widget_items.append({"pos" : (start_idx, jdx), "item" : table_widget_item})
    return table_widget_items

def update_table_rows(table, expiry_attribute, strikes_to_change, hide_flag, table_method):
    for strike in strikes_to_change:
        idx = expiry_attribute.strike_idx_map[strike]
        table.setRowHidden(idx, hide_flag)
        for opt_type in ["c", "p"]:
            opt_name = table.options_container.maps.expiry_strike_type_instrument_map[expiry_attribute.expiry][strike][opt_type]
            getattr(table.viewable_instruments, table_method)(opt_name)

@profile
def update_strikes_for_expiry(table, new_n_strikes, expiry_attribute):
    old_strikes = table.current_expiry_strike_map[expiry_attribute.expiry]
    new_strikes, _, _  = utils_omon.get_closest_n_strikes(table.strike_center, expiry_attribute.strikes, new_n_strikes)

    table.current_expiry_strike_map[expiry_attribute.expiry] = new_strikes

    if len(new_strikes) > len(old_strikes):
        strikes_to_change = [strike for strike in new_strikes if not strike in old_strikes]
        update_table_rows(table, expiry_attribute, strikes_to_change, False, "append")

    elif len(new_strikes) == len(old_strikes):
        strikes_to_add = [strike for strike in new_strikes if not strike in old_strikes]
        update_table_rows(table, expiry_attribute, strikes_to_add, False, "append")
        strikes_to_remove = [strike for strike in old_strikes if not strike in new_strikes]
        update_table_rows(table, expiry_attribute, strikes_to_remove, True, "remove")
    else:
        strikes_to_change = [strike for strike in old_strikes if not strike in new_strikes]
        update_table_rows(table, expiry_attribute, strikes_to_change, True, "remove")



from pyqtgraph.Qt import QtCore
from pyqtgraph.opengl.GLGraphicsItem import GLGraphicsItem
from pyqtgraph.Qt import QtCore, QtGui
from PySide6 import QtWidgets
from pyqtgraph.Qt import QT_LIB
import weakref
from pyqtgraph import opengl
import numpy as np
import warnings
#from .AbstractGLGraphicsItem import AbstractGLGraphicsItem, ABCGraphicsItemMeta

from .GL3DGraphicsItemMixin import GL3DGraphicsItemMixin

def get_item_data(item):
    if isinstance(item, opengl.GLScatterPlotItem):
        return item.pos
    elif isinstance(item, opengl.GLMeshItem):
        return item.opts['meshdata'].vertexes()
    elif isinstance(item, opengl.GLLinePlotItem):
        return item.pos
    elif isinstance(item, opengl.GLBarGraphItem):
        return item.points()
    return None



def compute_scene_bbox(items):
    min_vals, max_vals = None, None
    
    for item in items:
        data = get_item_data(item)
        if data is None or len(data) == 0:
            continue
            
        if not isinstance(data, np.ndarray):
            data = np.array(data)
            
        if data.ndim == 1:
            data = data[np.newaxis, :]
        if data.shape[1] == 2:
            data = np.hstack([data, np.zeros((len(data), 1))])
            
        item_min = np.nanmin(data, axis=0)
        item_max = np.nanmax(data, axis=0)
        
        if min_vals is None:
            min_vals, max_vals = item_min, item_max
        else:
            min_vals = np.minimum(min_vals, item_min)
            max_vals = np.maximum(max_vals, item_max)
    
    if min_vals is None:
        return np.array([0, 0, 0]), np.array([1, 1, 1])
    
    return min_vals, max_vals

class WeakList(object):

    def __init__(self):
        self._items = []

    def append(self, obj):
        #Add backwards to iterate backwards (to make iterating more efficient on removal).
        self._items.insert(0, weakref.ref(obj))

    def __iter__(self):
        i = len(self._items)-1
        while i >= 0:
            ref = self._items[i]
            d = ref()
            if d is None:
                del self._items[i]
            else:
                yield d
            i -= 1

#class GLItemGroup(AbstractGLGraphicsItem, QtWidgets.QGraphicsObject):

class GL3DItemGroup(GL3DGraphicsItemMixin):
    def __init__(self, parentItem=None):
        super().__init__(parentItem=parentItem)
        self.viewRange = [[0, 1] for _ in range(3)]
        self.itemsChangedListeners = WeakList()
        
        self.setFlag(GL3DGraphicsItemMixin.GLGraphicsItemFlag.ItemHasNoContents)
        self.setFlag(GL3DGraphicsItemMixin.GLGraphicsItemFlag.ItemSendsGeometryChanges)

    def itemChange(self, change, value):
        ret = super().itemChange(change, value)
        if change in [
            self.GLGraphicsItemChange.ItemChildAddedChange,
            self.GLGraphicsItemChange.ItemChildRemovedChange,
        ]:
            try:
                itemsChangedListeners = self.itemsChangedListeners
            except AttributeError:
                # It's possible that the attribute was already collected when the itemChange happened
                # (if it was triggered during the gc of the object).
                pass
            else:
                for listener in itemsChangedListeners:
                    listener.itemsChanged()

            try:
                inform_view_on_change = self.__inform_view_on_changes
            except AttributeError:
                # It's possible that the attribute was already collected when the itemChange happened
                # (if it was triggered during the gc of the object).
                pass
            else:
                if inform_view_on_change and change in [self.GLGraphicsItemChange.ItemPositionHasChanged, self.GLGraphicsItemChange.ItemTransformHasChanged]:
                    self.informViewBoundsChanged()
        return ret
        
    #def childItems(self):
    #    return list(self._GLGraphicsItem__children)

    def boundingRect(self):
        return QtCore.QRectF()
            
    def addItem(self, item):
        item.setParentItem(self)
        
    def _childPaint(self):...
    
    def _paintHelper(self):... 